/*
 * 本ソフトウェアは、MIT ライセンスの元で公開されています。
 * Copyright (c) 2022 重定 如彦
 * 以下の条件を満たす限り、自由な複製・配布・修正を無制限に許可する。
 * ・上記の著作権表示と本許諾書を、ソフトウェアの複製または重要な部分に記載する
 * ・本ソフトウェアは無保証である。自己責任で使用する。
 * 
 * Copyright (c) 2022 Yukihiko Shigesada
 * This software is released under the MIT License.
 * http://opensource.org/licenses/mit-license.php
 */

"use strict";
let showCode;

document.addEventListener('DOMContentLoaded', function () {
    new ShowCode();
});

const replace_data_list = [
    {
        name: "html|head|body|title|meta|h1|input|label|link|script|div|span|br|table|tr|td|canvas|img|a|p",
        regexp: "()(html|head|body|title|meta|h1|input|label|link|script|div|span|br|table|tr|td|canvas|img|a|p)(?=\\s*タグ)()",
        url: "html_tag"
    },
    {
        name: "id|class|charset|rel|href|src|border|type|value|name|checked|for",
        regexp: "()(id|class|charset|rel|href|src|border|type|value|name|checked|for)(?=\\s*属性)()",
        url: "html_attribute"
    },
    {
        name: "color|background-color|font-size|font|border-collapse|width|height|text-align|margin|margin-top|margin-right|margin-bottom|margin-left|user-select|background-image|background-size|border|border-style|border-width|border-color|box-sizing",
        regexp: "()(color|background-color|font-size|font|border-collapse|width|height|text-align|margin|margin-top|margin-right|margin-bottom|margin-left|user-select|background-image|background-size|border|border-style|border-width|border-color|box-sizing)(?=\\s*[プ=!」])()",
        url: "css_property"
    },
    {
        name: "break|continue",
        regexp: "([^a-zA-Z_\-])(break|continue)(?!\\s*[~文])([^a-zA-Z_\-])",
        url: "grammer_label"
    },
    {
        name: "case|default",
        regexp: "([^a-zA-Z_\-])(?<!upper\\s+|lower\\s+|camel\\s+)(case|default)([^a-zA-Z_\-])",
        url: "grammer_condition_switch"
    },
    {
        name: "debugger",
        regexp: "([^a-zA-Z_\-])(debugger)([^a-zA-Z_\-])",
        chapter: 4, dtype: "S", num: "11-4"
    },
    {
        name: "do",
        regexp: "([^a-zA-Z_\-])(do)(?!\\s*~)([^a-zA-Z_\-])",
        url: "grammer_loop_do_while"
    },
    {
        name: "if|else if|else",
        regexp: "([^a-zA-Z_\-])(if|else if|else)(?!\\s*[文~])([^a-zA-Z_\-])",
        url: "grammer_condition_if"
    },
    {
        name: "for",
        regexp: "([^a-zA-Z_\-])(for)(?!\\s*[~文])([^a-zA-Z_\-])",
        url: "grammer_loop_for"
    },
    {
        name: "this",
        regexp: "([^a-zA-Z_\-])(this)([^a-zA-Z_\-])",
        url: "grammer_function_this"
    },
    {
        name: "typeof",
        regexp: "([^a-zA-Z_\-])(typeof)([^a-zA-Z_\-])",
        url: "operator_unary"
    },
    {
        name: "while",
        regexp: "([^a-zA-Z_\-])(?<!~\\s*)(while)(?!\\s*文)([^a-zA-Z_\-])",
        url: "grammer_loop_while"
    },
    {
        name: "style|className|classList|innerText|innerHTML|value|firstChild|id",
        regexp: "([^a-zA-Z_\-])(style|className|classList|innerText|innerHTML|value|firstChild|id)(?=[\\s」]*[プ\.」])([^a-zA-Z_\-])",
        chapter: 8, dtype: "S", num: "4-1"
    },
    {
        name: "cssText",
        regexp: "([^a-zA-Z_\-])(cssText)([^a-zA-Z_\-])",
        chapter: 8, dtype: "S", num: "5"
    },
    {
        name: "@type|@param|@returns|@property|@typedef",
        regexp: "([^a-zA-Z_\-])(@type|@param|@returns|@property|@typedef)([^a-zA-Z_\-])",
        url: "jsdoc_tag"
    },
    {
        name: "keydown|keyup|click|mousedown|mouseup|mousemove|mouseenter|mouseleave|mouseout|contextmenu|load|change",
        regexp: "()(?<=on)(keydown|keyup|click|mousedown|mouseup|mousemove|mouseenter|mouseleave|mouseout|contextmenu|load|change)(?!=)([^a-zA-Z_\-])",
        url: "event"
    },
    {
        name: "keydown|keyup|click|mousedown|mouseup|mousemove|mouseenter|mouseleave|mouseout|contextmenu|load|change",
        regexp: "([^a-zA-Z_\-])(keydown|keyup|click|mousedown|mouseup|mousemove|mouseenter|mouseleave|mouseout|contextmenu|load|change)(?!=)([^a-zA-Z_\-])",
        url: "event"
    },
    { name: "コンソールオブジェクト|console.log|console.error", chapter: 6, dtype: "S", num: "7-1" },
    { name: "console.assert|console.clear", chapter: 6, dtype: "S", num: "7-2" },
    { name: "console.count|console.countReset", chapter: 6, dtype: "S", num: "7-3" },
    { name: "console.time|console.timeLog", chapter: 6, dtype: "S", num: "7-5" },
    {
        name: "Math",
        regexp: "()(Math)(?![a-zA-Z_\-])()",
        chapter: 6, dtype: "S", num: "8-1"
    },
    {
        name: "Math|PI|random|floor|ceil|trunc|round",
        regexp: "([^a-zA-Z_\-])(Math|PI|random|floor|ceil|trunc|round)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-1"
    },
    {
        name: "abs|sqrt|max|min|sin",
        regexp: "([^a-zA-Z_\-])(abs|sqrt|max|min|sin)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-2"
    },
    {
        name: "Number.isNaN|isNaN|Number.isFinite|isFinite|parseInt|parseFloat|String|alert|setTimeout|requestAnimationFrame",
        regexp: "([^a-zA-Z_\-])(Number.isNaN|isNaN|Number.isFinite|isFinite|parseInt|parseFloat|String|alert|setTimeout|requestAnimationFrame)(?!=)([^a-zA-Z_\-])",
        url: "buildin_function"
    },
    { name: "window", chapter: 8, dtype: "S", num: "2" },
    { name: "document|getElementById|createElement|addEventListener|DOMContentLoaded", chapter: 8, dtype: "S", num: "3-1" },
    {
        name: "key|code|stopPropagation",
        regexp: "([^a-zA-Z_\-])(key|code|stopPropagation)([^a-zA-Z_\-])",
        chapter: 8, dtype: "S", num: "6-2"
    },
    {
        name: "offsetX|offsetY|clientX|clientY|screenX|screenY|button|buttons|preventDefault",
        regexp: "([^a-zA-Z_\-])(offsetX|offsetY|clientX|clientY|screenX|screenY|buttons|button|preventDefault)([^a-zA-Z_\-])",
        chapter: 8, dtype: "S", num: "6-4"
    },
    {
        name: "toLowerCase|toUpperCase",
        regexp: "([^a-zA-Z_\-])(toLowerCase|toUpperCase)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-3"
    },
    {
        name: "replace|substr",
        regexp: "([^a-zA-Z_\-])(replace|substr)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-4"
    },
    {
        name: "length",
        regexp: "([^a-zA-Z_\-])(length)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-5"
    },
    {
        name: "isArray",
        regexp: "([^a-zA-Z_\-])(isArray)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-6"
    },
    {
        name: "Array|push|concat|indexOf|sort|entries",
        regexp: "([^a-zA-Z_\-])(?<!Object\\.)(Array|push|concat|indexOf|sort|entries)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-6"
    },
    {
        name: "getFullYear|getMonth|getDate|getHours|getMinutes|getSeconds|now",
        regexp: "([^a-zA-Z_\-])(getFullYear|getMonth|getDate|getHours|getMinutes|getSeconds|now)([^a-zA-Z_\-])",
        chapter: 6, dtype: "S", num: "8-8"
    },
    {
        name: "fillStyle|strokeStyle|lineWidth|font|beginPath|closePath|fill|stroke|rect|arc|moveTo|lineTo|fillText|clearRect|measureText|drawImage",
        regexp: "([^a-zA-Z_\-])(fillStyle|strokeStyle|lineWidth|font|beginPath|closePath|fill|stroke|rect|arc|moveTo|lineTo|fillText|clearRect|measureText|drawImage)([^a-zA-Z_\-])",
        url: "buildin_canvas"
    },
    { name: "appendChild|removeChild|getContext|focus", chapter: 8, dtype: "S", num: "4-2" },
    { name: "厳格モード", url: "grammer_strict" },
    { name: "let", regexp: "([^a-zA-Z_\-])(let)([^a-zA-Z_\-])", url: "grammer_variable" },
    { name: "var", regexp: "([^a-zA-Z_\-])(var)([^a-zA-Z_\-])", chapter: 2, dtype: "S", num: "3" },
    { name: "変数", regexp: "([^域|ル])(変数)([^名])", url: "grammer_variable" },
    { name: "const", url: "grammer_constant" },
    { name: "定数", regexp: "([^域|ル])(定数)([^名])", url: "grammer_constant" },
    { name: "function", regexp: "([^_])(function)([^\"])", url: "grammer_function" },
    { name: "Function", url: "grammer_function" },
    { name: "関数|メソッド", regexp: "([^み])(関数|メソッド)([^の名型呼])", url: "datatype_function" },
    { name: "関数呼び出し", url: "grammer_function_call" },
    { name: "関数の定義", url: "grammer_function" },
    { name: "引数|返り値", regexp: "([^余])(引数|返り値)()", url: "grammer_function" },
    { name: "return", regexp: "([^@])(return)()", url: "grammer_function_call" },
    { name: "データ型判定", url: "operator_unary" },
    { name: "データ型|プリミティブ値", regexp: "([])(データ型|プリミティブ値)([^判])", url: "datatype_index" },
    { name: "変数名|定数名|関数名|名前の制限", "url": "grammer_name" },
    { name: "大域変数|大域定数|グローバル変数|スコープ|衝突", "url": "grammer_scope" },
    { name: "ブロック", regexp: "([^言])(?<!ゲーム盤の|大きさの|マスに|構成する|複数の|ポリオミノ」?の|配置された|配置する|配置した|[赤青]い|[緑色]の)(ブロック)(?!の形|の一部|の表示|タグ|を削除|.配置|と重なる|がゲーム|の数|の番号|を画面|を上|を表示|が存在)()", "url": "grammer_block" },
    { name: "予約語", "url": "grammer_reserved" },
    { name: "代入文|代入", regexp: "()(代入文|代入)([^演])", "url": "grammer_assignment" },
    { name: "コメント", regexp: "([^の])(コメント)()", "url": "grammer_comment" },
    { name: "Javascript のコメント", "url": "grammer_comment" },
    { name: "if ~ else 文|if 文", regexp: "()(?<!~\\s*)(if\\s*文|if\\s*~\\s*else\\s*文)()", "url": "grammer_condition_if" },
    { name: "switch 文|switch", regexp: "([^a-zA-Z_\-])(?<!~\\s*)(switch\\s*文|switch)([^a-zA-Z_\-])", "url": "grammer_condition_switch" },
    { name: "繰り返し", "url": "grammer_index" },
    { name: "while 文", regexp: "()(?<!~\\s*)(while\\s*文)()", "url": "grammer_loop_while" },
    { name: "do ~ while 文", "url": "grammer_loop_do_while" },
    { name: "for 文", regexp: "()(?<!~\\s*)(for\\s*文)()", "url": "grammer_loop_for" },
    { name: "for ~ in 文", regexp: "()(for\\s*~\\s*in\\s*文)()", "url": "grammer_loop_for_in" },
    { name: "for ~ of 文", regexp: "()(for\\s*~\\s*of\\s*文)()", "url": "grammer_loop_for_of" },
    { name: "break 文", regexp: "()(?<!~\\s*)(break\\s*文)()", "url": "grammer_label" },
    { name: "continue 文", regexp: "()(?<!~\\s*)(continue\\s*文)()", "url": "grammer_label" },
    { name: "falsy|truthy", url: "grammer_tandf" },
    { name: "ラベル", "url": "grammer_label" },
    { name: "数値型|数値|Infinity", url: "datatype_number" },
    { name: "NaN", regexp: "([^a-zA-Z_\-])(NaN)([^a-zA-Z_\-])", url: "datatype_number" },
    { name: "文字列型|文字列|エスケープ文字|エスケープ表記", url: "datatype_string" },
    { name: "Boolean 型|true|false|論理型", url: "datatype_boolean" },
    { name: "Undefined 型|未定義型", url: "datatype_undefined" },
    { name: "undefined", regexp: "([^_])(undefined)([^\"])", url: "datatype_undefined" },
    { name: "Null 型", url: "datatype_null" },
    { name: "null", regexp: "([^_])(null)(?! 合体)([^\"])", url: "datatype_null" },
    { name: "配列", url: "datatype_array" },
    { name: "関数型", url: "datatype_function" },
    { name: "Math オブジェクト", chapter: 6, dtype: "S", num: "8-1" },
    { name: "String オブジェクト", chapter: 6, dtype: "S", num: "8-3" },
    { name: "ウィンドウオブジェクト", chapter: 8, dtype: "S", num: "2" },
    { name: "要素オブジェクト", chapter: 8, dtype: "S", num: "4-1" },
    { name: "ドキュメントオブジェクト", chapter: 8, dtype: "S", num: "3-1" },
    { name: "CSS 宣言ブロックオブジェクト", chapter: 8, dtype: "S", num: "5" },
    { name: "イベントオブジェクト", chapter: 8, dtype: "S", num: "6-1" },
    { name: "Date オブジェクト|Date()", chapter: 6, dtype: "S", num: "8-7" },
    { name: "オブジェクト", regexp: "([^みト\\s素ウクル])(オブジェクト)()", url: "datatype_object" },
    { name: "プロパティ", url: "datatype_object" },
    { name: "演算子", regexp: "([^術入理較項価件ト算余])(演算子)()", url: "operator_index" },
    { name: "算術演算子|オペランド", url: "operator_arithmetic" },
    { name: "加算演算子|減算演算子|乗算演算子|除算演算子|剰余演算子|べき乗演算子|加算|減算|乗算|除算|剰余|べき乗", url: "operator_arithmetic", only: "operator_precedence" },
    { name: "比較演算子", url: "operator_conditional" },
    { name: "小なり|大なり|厳密等価演算子|厳密等価|等価演算子|等価|厳密不等価演算子|不等価演算子|厳密不等価|不等価", url: "operator_conditional", only: "operator_precedence" },
    { name: "代入演算子", url: "operator_assignment" },
    { name: "前置インクリメント演算子|後置インクリメント演算子|前置デクリメント演算子|後置デクリメント演算子|インクリメント演算子|デクリメント演算子|前置インクリメント|後置インクリメント|前置デクリメント|後置デクリメント|インクリメント|デクリメント", url: "operator_assignment", only: "operator_precedence" },
    { name: "null 合体", url: "operator_logical" },
    { name: "論理演算子|短絡", url: "operator_logical" },
    { name: "論理積|論理和|論理否定|AND|OR|NOT", url: "operator_logical", only: "operator_precedence" },
    { name: "単項演算子", url: "operator_unary" },
    { name: "単項正値|単項負値|データ型判定", url: "operator_unary" },
    { name: "その他の演算子", url: "operator_other" },
    { name: "グループ化|条件演算子", url: "operator_other", only: "operator_precedence" },
    { name: "優先順位", url: "operator_precedence" },
    { name: "暗黙の型変換", url: "operator_implicit_conversion" },
    { name: "明示的な型変換", url: "operator_explicit_conversion" },
    { name: "組み込み関数", url: "buildin_function" },
    { name: "組み込みオブジェクト", url: "buildin_index" },
    { name: "イベント", regexp: "()(イベント)([^オ])", url: "event" },
    { name: "HTML のコメント", url: "html_grammer" },
    { name: "CSS のコメント", url: "css_grammer" },
    { name: "属性", url: "html_attribute" },
    { name: "JSDoc", url: "jsdoc_index" },
    { name: "VSCode", url: "vscode_display" },
    { name: "条件分岐", url: "grammer_index" },
    { name: "文法エラー", url: "error_syntax" },
    { name: "実行エラー", url: "error_execute" },
    { name: "論理エラー", url: "error_logic" },
    { name: "エラーメッセージ|エラー", regexp: "([^法行理])(エラーメッセージ|エラー)()", url: "error_index" },
    //{ name: "++|--", regexp: "()(\\+\\+|\\-\\-)()", url: "operator_assignment" },
    { name: "ブロックタグ", url: "jsdoc_tag" },
    { name: "ソース", url: "dtool_source" },
    { name: "コンソール", regexp: "()(コンソール)([^オ])", url: "dtool_console" },
    { name: "セレクタ|トグルボタン|教材の選択ボタン", url: "usage_buttontype" },
    { name: "戻る、進むボタン|戻るボタン|進むボタン", url: "usage_move" },
    { name: "章の選択ボタン", url: "usage_chapter" },
    { name: "資料の表示ボタン|資料画面", url: "usage_appendix" },
    { name: "検索の設定", url: "usage_search_config" },
    { name: "検索ボタン|検索画面", url: "usage_search" },
    { name: "設定ボタン|設定画面", url: "usage_config" },
    { name: "ショートカットキーボタン|ショートカットキー", url: "usage_shortcutkey" },
    { name: "ヘルプボタン|ヘルプ", url: "usage_help" },
    { name: "教材の選択ボタン", url: "usage_material" },
    { name: "説明文", url: "usage_description" },
    { name: "実行結果", url: "usage_execute" },
    { name: "プログラムの操作ボタン", url: "usage_source_command" },
    { name: "プログラムの表示", url: "usage_source" },
    { name: "ブレークポイント", chapter: 4, dtype: "S", num: "11-5", },
    { name: "ステップイン", chapter: 4, dtype: "S", num: "11-7", },
    { name: "ステップオーバー", chapter: 4, dtype: "S", num: "11-8", },
    { name: "ステップアウト", chapter: 4, dtype: "S", num: "11-9", },
];

class ShowCode {
    constructor(reset = false) {
        showCode = this;
        this.param = {};
        //this.param.chapter = "2章";
        this.param.diff_context = 5;
        this.setparam({
            left: null,
            leftname: "",
            right: null,
            rightname: "",
            diff_text: null,
            show_execute_src: null,
            show_execute: false,
            show_diff_textarea: false,
            delete_comment: false,
            execute: true,
        });

        $("#page_prev").click(() => {
            let history = this.param.page_history[--this.param.page_history_index];
            this.param.chapter = history.chapter;
            let data = JSON.parse(history.data);
            this.change_page(data, null, false);
            this.param.change_page = {
                name: this.calc_name(data),
                iscode: false,
                keyword: "",
            };
            this.redraw();
        });
        $("#page_next").click(() => {
            let history = this.param.page_history[++this.param.page_history_index];
            this.param.chapter = history.chapter;
            let data = JSON.parse(history.data);
            this.change_page(data, null, false);
            this.param.change_page = {
                name: this.calc_name(data),
                iscode: false,
                keyword: "",
            };
            this.redraw();
        });
        for (let chapter in chapter_data) {
            let data = chapter_data[chapter];
            if (data.length === 0) {
                continue;
            }
            $(`#${chapter}`).click(() => {
                $("#index").scrollTop(0);
                this.change_page(chapter_data[chapter][0], chapter);
            });
        }
        $("#appendix_button").click(() => {
            if (!this.param.appendix_hidden) {
                this.hide_appendix();
            }
            else {
                this.param.appendix_hidden = false;
            }
            if (!this.param.appendix || !this.param.appendix_history) {
                this.param.appendix = "index.html";
                this.param.appendix_history_index = 0;
                this.param.appendix_history = [this.param.appendix];
            }
            if (!this.param.appendix_hidden) {
                this.param.search.hidden = true;
                this.param.config.hidden = true;
            }
            this.redraw(false);
        });
        $("#shortcut_button").click(() => {
            this.param.shortcut = !this.param.shortcut;
            if (this.param.shortcut) {
                $("#shortcut_button").addClass("selected");
            }
            else {
                $("#shortcut_button").removeClass("selected");
            }
            this.help.sethelp();
        });
        $("#help_button").click(() => {
            this.param.help = !this.param.help;
            if (this.param.help) {
                $("#help_button").addClass("selected");
            }
            else {
                $("#help_button").removeClass("selected");
                $("#help").addClass("hidden");
            }
            this.help.sethelp();
        });
        $("#search_button").click(() => {
            this.param.search.hidden = !this.param.search.hidden;
            if (this.param.search.result === undefined) {
                this.param.search.result = "";
            }
            if (!this.param.search.hidden) {
                this.hide_appendix();
                this.param.config.hidden = true;
            }
            this.redraw(false);
        });
        $("#config_button").click(() => {
            this.param.config.hidden = !this.param.config.hidden;
            if (!this.param.config.hidden) {
                this.hide_appendix();
                this.param.search.hidden = true;
            }
            this.redraw(false);
        });
        $("#show_execute").click(() => {
            this.param.execute_min = false;
            this.redraw();
        });
        $("#show_description").click(() => {
            this.param.desc_min = false;
            this.redraw();
        });

        $("#HTML_selector").click(() => {
            this.setparam({
                left: null,
                leftname: "",
                right: this.param.execute_html,
                rightname: "",
                show_execute_src: "HTML",
                show_execute: true,
                show_diff_textarea: false,
                diff_text: null,
            })
            this.redraw();
        });
        $("#CSS_selector").click(() => {
            this.setparam({
                left: null,
                leftname: "",
                right: this.param.execute_css,
                rightname: "",
                show_execute_src: "CSS",
                show_execute: true,
                show_diff_textarea: false,
                diff_text: null,
            })
            this.redraw();
        });
        $("#Javascript_selector").click(() => {
            this.setparam({
                left: null,
                leftname: "",
                right: this.param.execute_script,
                rightname: "",
                show_execute_src: "Javascript",
                show_execute: true,
                show_diff_textarea: false,
                diff_text: null,
            })
            this.redraw();
        });
        //$("#stop_execute").click(() => {
        //    this.param.show_execute = false;
        //    this.redraw();
        //});
        $("#diff").click(() => {
            if (!this.param.left || (this.param.data.pfname && this.param.left !== this.param.data.pfname)) {
                let data = this.param.data;
                let pname = this.find_and_calc_name(data);
                this.param.left = data.pfname;
                this.param.leftname = pname;
                this.param.right = data.fname;
                this.param.pfname = data.pfname;
                this.param.diff_text = null;
                this.show_execute(data);
                this.redraw();
            }
            else {
                this.param.left = null;
                this.param.leftname = "";
                this.param.show_diff_textarea = false;
                this.param.diff_text = null;
                this.redraw();
            }
        });

        $("#start_execute").click(() => {
            this.save_and_reload();
        });
        $("#debug").click(() => {
            this.param.debug = true;
            this.save_and_reload();
        });
        $("#copy").click(() => {
            let fname = this.calc_fname(this.param.right);
            $("#textarea").removeClass("hidden");
            $("#textarea").text(filedata[fname]).select();
            document.execCommand("Copy");
            $("#textarea").addClass("hidden");
        });
        $("#diff_box").click(() => {
            this.param.left = null;
            this.param.leftname = "";
            this.param.diff_text = null;
            this.param.show_diff_textarea = !this.param.show_diff_textarea;
            this.redraw();
        });
        $("#diff_file").click(() => {
            let file_elem = $('<input type="file" id="file">');
            file_elem.click();
            file_elem.on("change", (evt) => {
                if (file_elem[0].files.length === 0) {
                    return;
                }
                this.process_diff_files(file_elem[0].files);
            })
        });
        $("#diff_clipboard").click(() => {
            navigator.clipboard.readText().then(
                (clipText) => {
                    this.param.diff_text = clipText;
                    this.param.show_diff_textarea = false;
                    this.param.left = null;
                    this.param.leftname = "クリップボード";
                    if (!this.param.diff_text) {
                        alert("比較するクリップボードの中身が空か、テキストデータではありません。")
                    }
                    this.redraw();
                }
            );
            return;
        });
        $("#comment").click(() => {
            this.param.delete_comment = !this.param.delete_comment;
            this.redraw();
        });
        $("#desc_minmax").click(() => {
            this.param.desc_min = !this.param.desc_min;
            //this.redrawflag = false;
            this.redraw();
        });
        $("#desc_prev").click(() => {
            [this.param.description, this.param.desc] = this.param.desc_history.pop();
            this.set_description_style();
            this.set_index_style();
            this.save_storage();
        });
        $("#execute_minmax").click(() => {
            this.param.execute_min = !this.param.execute_min;
            //this.redrawflag = false;
            this.redraw();
        });
        $("#appendix_prev").click(() => {
            this.param.appendix = this.param.appendix_history[--this.param.appendix_history_index];
            this.redraw(false);
        });
        $("#appendix_next").click(() => {
            this.param.appendix = this.param.appendix_history[++this.param.appendix_history_index];
            this.redraw(false);
        });
        $("#appendix_min").click(() => {
            this.hide_appendix();
            this.redraw(false);
        });
        $("#appendix_close").click(() => {
            this.hide_appendix();
            this.param.appendix_scrollTop = 0;
            this.param.appendix = "index.html";
            this.param.appendix_history_index = 0;
            this.param.appendix_history = [this.param.appendix];
            this.redraw(false);
        });
        //$("#start_search").click(() => {
        //    this.search();
        //});
        $("#search_option_button").click(() => {
            if (!this.param.search.hidden) {
                this.param.search.option = !this.param.search.option;
                this.redraw(false);
            }
        });
        $("#search_text").change(() => {
            this.param.search.option = false;
            this.search(false);
        });
        $("#search_text").keydown((evt) => {
            if (evt.key === "Enter") {
                this.param.search.option = false;
                this.search(false);
            }
        });
        $("#search_min").click(() => {
            this.param.search.hidden = true;
            this.redraw(false);
        });
        $("#search_close").click(() => {
            this.param.search.hidden = true;
            this.redraw(false);
        });
        $("#config_close").click(() => {
            this.param.config.hidden = true;
            this.redraw(false);
        });

        $("*").on("dragover", (evt) => {
            evt.preventDefault();
        });
        $("*").on("drop", (evt) => {
            if (evt.originalEvent.dataTransfer.files.length === 0) {
                return;
            }
            evt.stopPropagation();
            evt.preventDefault();
            if (this.param.right === null) {
                return;
            }
            this.process_diff_files(evt.originalEvent.dataTransfer.files);
        });

        $(window).resize(() => this.redraw());//set_index_style());
        if (!reset) {
            this.load_storage();
        }

        this.set_search_handler();
        // set_search_handler の中で、空白を取り除いて作り直される場合があるので、
        // search に関するイベントハンドラはこの後で定義する必要がある
        $("#search_material_chapter_all").click(() => {
            for (let c = 2; c <= 15; c++) {
                this.param.search.material_chapter[c] = true;
            }
            this.set_search_class();
        });
        $("#search_material_chapter_clear").click(() => {
            for (let c = 2; c <= 15; c++) {
                this.param.search.material_chapter[c] = false;
            }
            this.set_search_class();
        });
        $("#search_material_chapter_prev").click(() => {
            for (let c = 15; c >= 2; c--) {
                if (this.param.search.material_chapter[c]) {
                    for (let c2 = 2; c2 < c; c2++) {
                        this.param.search.material_chapter[c2] = true;
                    }
                    break;
                }
            }
            this.set_search_class();
        });
        $("#search_material_chapter_after").click(() => {
            for (let c = 2; c <= 15; c++) {
                if (this.param.search.material_chapter[c]) {
                    for (let c2 = c; c2 <= 15; c2++) {
                        this.param.search.material_chapter[c2] = true;
                    }
                    break;
                }
            }
            this.set_search_class();
        });
        $("#search_material_type_all").click(() => {
            this.param.search.material_type.main = true;
            this.param.search.material_type.misc = true;
            this.param.search.material_type.sup = true;
            this.param.search.material_type.question = true;
            this.set_search_class();
        });
        $("#search_material_type_clear").click(() => {
            this.param.search.material_type.main = false;
            this.param.search.material_type.misc = false;
            this.param.search.material_type.sup = false;
            this.param.search.material_type.question = false;
            this.set_search_class();
        });
        $("#search_load_default").click(() => {
            if (confirm(`検索設定を初期設定に戻します`)) {
                this.param.search = {};
                this.set_search_handler();
                this.set_search_class();
                this.param.search.option = true;
                this.save_and_reload();
            }
        });
        for (let i = 1; i <= 5; i++) {
            $(`#search_load_setting${i}`).click(() => {
                if (confirm(`検索設定を「設定${i}」から読み込みます`)) {
                    this.param.search_save = this.param.search_save || [];
                    this.param.search_save[i] = this.param.search_save[i] || {};
                    this.param.search = { ...this.param.search_save[i] };
                    this.param.search.option = true;
                    this.set_search_handler();
                    this.set_search_class();
                    this.save_and_reload();
                }
            });
            $(`#search_save_setting${i}`).click(() => {
                if (confirm(`現在の検索設定を「設定${i}」に保存します`)) {
                    this.param.search_save = this.param.search_save || [];
                    this.param.search_save[i] = { ...this.param.search };
                    this.save_and_reload();
                }
            });
        }
        $("#delete_history").click(() => {
            if (confirm(`教材の表示の履歴を削除します`)) {
                this.param.page_history = [{ chapter: this.param.chapter, data: JSON.stringify(this.param.data) }];
                this.param.page_history_index = 0;
                this.redraw();
            }
        });
        $("#config_default").click(() => {
            if (confirm(`設定を初期設定に戻します`)) {
                for (let param of this.config_params) {
                    this.param.config[param.name] = param.default;
                }
                this.redraw();
            }
        });
        this.set_config_handler();

        this.param.execute = true;
        this.param.search = this.param.search || {};
        this.param.config = this.param.config || {};
        if (!this.param.chapter || !chapter_data[this.param.chapter]) {
            this.param.chapter = "2章";
            this.param.page_history = [];
            this.param.page_history_index = -1;
            this.change_page(chapter_data[this.param.chapter][0], null, true, false);
        }
        if (!this.param.page_history) {
            this.param.page_history = [{ chapter: this.param.chapter, data: JSON.stringify(this.param.data) }];
            this.param.page_history_index = 0;
        }
        if (!this.param.appendix_history) {
            this.param.appendix = "index.html";
            this.param.appendix_history_index = 0;
            this.param.appendix_history = [this.param.appendix];
            this.param.appendix_scrollTop = 0;
            this.hide_appendix();
        }

        this.help = new Help(this);

        this.animation();
        this.redraw();
    }

    search_params = [
        { name: "material", type: "boolean", default: true },
        { name: "material_type", object: true, default: true, disable: "material" },
        { name: "material_description", type: "string", default: "all", disable: "material" },
        { name: "material_code", type: "string", default: "all", disable: "material" },
        { name: "material_code_line", object: true, default: true, disable: "material" },
        { name: "material_code_word", object: true, default: true, disable: "material" },
        { name: "material_chapter", object: true, default: true, disable: "material", width: 22, nospace: true },
        { name: "appendix", type: "string", default: "all" },
    ];

    config_params = [
        { name: "button_disptype_chapter", type: "string", default: "normal" },
        { name: "button_disptype_appendix", type: "string", default: "normal" },
        { name: "button_disptype_description", type: "string", default: "normal" },
        { name: "button_disptype_lang", type: "string", default: "normal" },
        { name: "button_disptype_execute", type: "string", default: "normal" },
        { name: "button_disptype_diff", type: "string", default: "normal" },
        { name: "description_maxheight", type: "string", default: "50" },
    ];

    button_texts = {
        chapter: [
            { id: "章", normal: "", simple: "章 " },
            { id: "2章", normal: "2章", simple: "２" },
            { id: "3章", normal: "3章", simple: "３" },
            { id: "4章", normal: "4章", simple: "４" },
            { id: "5章", normal: "5章", simple: "５" },
            { id: "6章", normal: "6章", simple: "６" },
            { id: "7章", normal: "7章", simple: "７" },
            { id: "8章", normal: "8章", simple: "８" },
            { id: "9章", normal: "9章", simple: "９" },
            { id: "10章", normal: "10章", simple: "10" },
            { id: "11章", normal: "11章", simple: "11" },
            { id: "12章", normal: "12章", simple: "12" },
            { id: "13章", normal: "13章", simple: "13" },
            { id: "14章", normal: "14章", simple: "14" },
            { id: "15章", normal: "15章", simple: "15" },
        ],
        appendix: [
            { id: "appendix_button", normal: "資料", simple: "資" },
            { id: "search_button", normal: "検索", simple: "検" },
            { id: "config_button", normal: "設定", simple: "設" },
        ],
        description: [
            { id: "show_execute", normal: "実行結果", simple: "実" },
            { id: "show_description", normal: "説明文", simple: "説" },
        ],
        lang: [
            { id: "HTML_selector", normal: "HTML", simple: "Ｈ" },
            { id: "CSS_selector", normal: "CSS", simple: "Ｃ" },
            { id: "Javascript_selector", normal: "JS", simple: "Ｊ" },
        ],
        execute: [
            { id: "start_execute", normal: "再実行", simple: "実行", letter: "実" },
            { id: "debug", normal: "デバッグ", simple: "ﾃﾞﾊﾞｯｸﾞ", letter: "デ" },
        ],
        diff: [
            { id: "diff", normal: "比較", simple: "比較", letter: "比" },
            { id: "comment", normal: "コメント", simple: "ｺﾒﾝﾄ", letter: "注" },
            { id: "diff_box", normal: "比較ボックス", simple: "ﾎﾞｯｸｽ", letter: "ボ" },
            { id: "diff_clipboard", normal: "クリップボードと比較", simple: "ｸﾘｯﾌﾟ", letter: "ク" },
            { id: "diff_file", normal: "ファイルと比較", simple: "ﾌｧｲﾙ", letter: "フ" },
            { id: "copy", normal: "コピー", simple: "ｺﾋﾟｰ", letter: "コ" },
        ],
    };

    set_search_handler() {
        this.set_handler("search", this.search_params);
    }

    set_config_handler() {
        this.set_handler("config", this.config_params);
    }

    set_handler(type, params) {
        this.param[type] = this.param[type] || {};
        for (const param of params) {
            const id = `${type}_${param.name}`;
            const regid = id.replace(/_/g, "\\_");
            let container = $(`#${id}`);
            if (param.nospace) {
                container.html(container.html().replace(/ *\n */g, ""));
            }
            if (this.param[type][param.name] === undefined) {
                if (!param.object) {
                    this.param[type][param.name] = param.default;
                }
                else {
                    this.param[type][param.name] = {};
                }
            }
            switch (param.type) {
                case "boolean":
                    this.param[type][param.name] = !!this.param[type][param.name];
                    break;
            }
            for (let selector of container.children()) {
                if (!param.object && selector.classList.contains("selector")) {
                    let match = selector.id.match(RegExp(`${regid}\_(.*)`));
                    switch (param.type) {
                        case "boolean":
                            match[1] = (match[1] === "true");
                            break;
                    }
                    selector.onclick = (evt) => {
                        this.param[type][param.name] = match[1];
                        this.set_class(type, params);
                        if (type === "config") {
                            this.redraw();
                        }
                    };
                }
                else if (param.object && selector.classList.contains("toggle_button")) {
                    let match = selector.id.match(RegExp(`${regid}\_(.*)`));
                    if (this.param[type][param.name][match[1]] === undefined) {
                        this.param[type][param.name][match[1]] = param.default;
                    }
                    selector.onclick = (evt) => {
                        this.param[type][param.name][match[1]] = !this.param[type][param.name][match[1]];
                        this.set_class(type, params);
                        if (type === "config") {
                            this.redraw();
                        }
                    };
                }
            }
        }
    }

    set_search_class() {
        this.set_class("search", this.search_params);
    }

    set_config_class() {
        this.set_class("config", this.config_params);
    }

    set_class(type, params) {
        for (const param of params) {
            const id = `${type}_${param.name}`;
            const regid = id.replace(/_/g, "\\_");
            const container = $(`#${id}`);
            for (let selector of container.children()) {
                let isselector = selector.classList.contains("selector");
                let isbutton = selector.classList.contains("toggle_button");
                if (isselector || isbutton) {
                    let match = selector.id.match(RegExp(`${regid}\_(.*)`));
                    if (match) {
                        selector.classList.remove("selected");
                        selector.classList.remove("not_selected");
                        selector.classList.remove("disabled_selected");
                        selector.classList.remove("disabled_not_selected");
                        let disabled = param.disable && !this.param[type][param.disable];
                        let isselected;
                        if (!param.object && isselector) {
                            isselected = (match[1] === "" + this.param[type][param.name]);
                        }
                        else if (param.object && isbutton) {
                            isselected = this.param[type][param.name][match[1]];
                        }
                        else {
                            continue;
                        }
                        if (isselected) {
                            if (!disabled) {
                                selector.classList.add("selected");
                            }
                            else {
                                selector.classList.add("disabled_selected");
                            }
                        }
                        else {
                            if (!disabled) {
                                selector.classList.add("not_selected");
                            }
                            else {
                                selector.classList.add("disabled_not_selected");
                            }
                        }
                        if (param.width) {
                            selector.style.width = param.width + "px";
                        }
                    }
                }
            }
        }
    }

    set_button_texts() {
        for (let name in this.button_texts) {
            for (let data of this.button_texts[name]) {
                $(`#${data.id}`).text(data[this.param.config["button_disptype_" + name]]);
            }
        }
    }

    animation() {
        let execute_height = $("#execute").height();
        //if (execute_height !== this.execute_height) {
        //this.set_description_style();
        this.execute_height = execute_height;
        let description_height = $("#description").get(0).scrollHeight;
        //if (this.redrawflag || this.imgloaded) {
        if (this.description_height !== description_height) {
            this.redrawflag = true;
            this.set_description_style();
        }
        this.description_height = description_height;

        let appendix_height = $("#appendix_contents").get(0).scrollHeight;
        if (!this.param.appendix_hidden && this.appendix_height !== appendix_height) {
            this.redrawflag = true;
            this.set_appendix_style();
        }
        this.appendix_height = appendix_height;

        let search_height = $("#search_result").get(0).scrollHeight;
        if (!this.param.search.hidden && this.search_height !== search_height) {
            this.redrawflag = true;
            this.create_search();
            $("#search_text").focus();
        }
        this.search_height = search_height;

        this.diff_set_scrollbar();
        this.set_index_style();
        requestAnimationFrame(this.animation.bind(this));
    }

    calc_fname(name, chapter) {
        chapter = chapter || this.param.chapter;
        return `..\\${chapter}\\${name}`
    }

    setparam(param) {
        param = param || {};
        for (let prop in param) {
            this.param[prop] = param[prop];
        }
    }

    redraw(setstyle = true) {
        $("input").off("keydown.showCode");
        $("textarea").off("keydown.showCode");
        $("input").off("keyup.showCode");
        $("textarea").off("keyup.showCode");

        this.redrawflag = false;
        if (this.param.execute) {
            if (this.param.change_page) {
                if (this.param.change_page.iscode) {
                    this.param.left = null;
                    this.param.show_execute_src = this.param.change_page.iscode;
                    if (this.param.show_execute_src === "JS") {
                        this.param.show_execute_src = "Javascript";
                        this.param.right = this.param.execute_script;
                    }
                    else if (this.param.show_execute_src === "CSS") {
                        this.param.right = this.param.execute_css;
                    }
                    else {
                        this.param.right = this.param.execute_html;
                    }
                    this.param.desc_min = true;
                    this.param.execute_min = true;
                }
            }
        }
        this.set_button_texts();
        this.create_chapter();
        this.create_appendix();
        this.create_search();
        this.create_config();
        this.create_index();
        if (setstyle) {
            this.create_diff();
            this.set_execute_style();
            this.set_description_style();
        }
        this.help.sethelp();
        if (this.param.execute) {
            this.param.execute = false;
            // this.execute 内でエラーになる可能性があるので、先に save_storage を実行しておく
            this.save_storage();
            this.set_index_style();
            if (!this.param.change_page) {
                $("#index").scrollTop(this.param.index_scrolltop);
            }
            else {
                $("#index").scrollTop($(`#index_${this.param.change_page.name}`).offset().top - $("#index_heading").offset().top);
                delete this.param.change_page;
            }
            this.execute();
            this.set_index_style();
        }
        else {
            this.save_storage();
            this.set_index_style();
        }

        $("input").on("keydown.showCode", function (evt) {
            if (this.type === "text") {
                evt.stopPropagation();
            }
        });
        $("textarea").on("keydown.showCode", function (evt) {
            evt.stopPropagation();
        });
        $("input").on("keyup.showCode", function (evt) {
            if (this.type === "text") {
                evt.stopPropagation();
            }
        });
        $("textarea").on("keyup.showCode", function (evt) {
            evt.stopPropagation();
        });
    }

    create_chapter() {
        for (let chapter in chapter_data) {
            let data = chapter_data[chapter];
            if (data.length === 0) {
                continue;
            }
            if (chapter === this.param.chapter) {
                $(`#${chapter}`).addClass("selected");
            }
            else {
                $(`#${chapter}`).removeClass("selected");
            }
        }
        if (this.param.shortcut) {
            $("#shortcut_button").addClass("selected");
        }
        else {
            $("#shortcut_button").removeClass("selected");
        }
        if (this.param.help) {
            $("#help_button").addClass("selected");
        }
        else {
            $("#help_button").removeClass("selected");
        }
    }

    create_appendix() {
        $("#appendix_prev").addClass("invisible");
        $("#appendix_next").addClass("invisible");
        if (!this.param.appendix_hidden) {
            $("#appendix_button").addClass("selected");
            $("#appendix").removeClass("hidden");
            $("#appendix_contents").empty();
            let fdata = filedata["..\\資料\\" + this.param.appendix];
            fdata = this.calc_appendix(fdata);

            let match = fdata.match(/(<h1>.+<\/h1>)/);
            if (match) {
                $("#appendix_title").html(match[1]);
                fdata = fdata.replace(/(<h1>.+<\/h1>)/, "");
            }
            else {
                $("#appendix_title").html("");
            }
            $("#appendix_contents").append(fdata);
            if (this.param.appendix_history_index > 0) {
                $("#appendix_prev").removeClass("invisible");
            }
            if (this.param.appendix_history_index < this.param.appendix_history.length - 1) {
                $("#appendix_next").removeClass("invisible");
            }
            $("#appendix_contents").height("auto");
            this.set_appendix_style();
        }
        else {
            $("#appendix_button").removeClass("selected");
            $("#appendix").addClass("hidden");
        }
    }

    hide_appendix() {
        if (!this.param.appendix_hidden) {
            this.param.appendix_scrollTop = $("#appendix_contents").scrollTop();
        }
        this.param.appendix_hidden = true;
    }

    calc_appendix(fdata) {
        let linkdata = [];
        let match;

        match = fdata.match(/(<h1>.+＞)([^<＞]+)(<\/h1>)/);
        if (match) {
            fdata = fdata.replace(/(<h1>.+＞)([^<＞]+)(<\/h1>)/, `$1!linkdata${linkdata.length}!$3`);
            linkdata.push(`${match[2]}`)
        }

        while (match = fdata.match(/!!([^!,]+),([^,!]+)!!/)) {
            fdata = fdata.replace(/!!([^!,]+),([^,!]+)!!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="#" onclick="change_appendix('${match[1]}')">${match[2]}</a>`)
        }
        //let linkdata2 = [];
        //count = 0;
        while (match = fdata.match(/!\?([^!,]+),([^,!]*),([^,!]*),([^,!]+),([^,!]+)!\?/)) {
            let num = 0;
            for (let data of chapter_data[match[1]]) {
                if (data.type === match[2] && data.dtype === match[3] && data.num === match[4]) {
                    break;
                }
                num++;
            }
            fdata = fdata.replace(/!\?([^!,]+),([^,!]*),([^,!]*),([^,!]+),([^,!]+)!\?/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="#" onclick="change_page('${match[1]}', '${num}')" class="material">${match[5]}</a>`)
        }

        //fdata = fdata.replace(/!!([^,]+),([^,!]+)!!/g, `<a href="#" onclick="change_appendix('$1')">$2</a>`);

        // 外部のリンク
        while (match = fdata.match(/!=http([^!,]+),([^!,]+)=!/)) {
            fdata = fdata.replace(/!=http([^!,]+),([^!,]+)=!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="http${match[1]}" class="external" target="blank">${match[2]}</a>`);
        }

        while (match = fdata.match(/!=([^!,]+)=!/)) {
            fdata = fdata.replace(/!=([^!,]+)=!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="https://developer.mozilla.org/ja/docs/${match[1]}" class="external" target="blank">参考ページ</a>`);
        }

        while (match = fdata.match(/!=([^!,]+),([^!,]+)=!/)) {
            fdata = fdata.replace(/!=([^!,]+),([^!,]+)=!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="https://developer.mozilla.org/ja/docs/${match[1]}" class="external" target="blank">${match[2]}</a>`);
        }

        while (match = fdata.match(/!-([^!,]+),([^!,]+)-!/)) {
            fdata = fdata.replace(/!-([^!,]+),([^!,]+)-!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="https://developer.mozilla.org/ja/docs/${match[1]}" class="external no_underline" target="blank">${match[2]}</a>`);
        }

        let re = new RegExp(/!\?(\d+,)?(HTML|CSS|JS|例|補足|章末問題|ヒント|解答例)([^,!]+)(,[^,!]*)?!\?/);
        while (match = fdata.match(re)) {
            let num = 0;
            let match2;
            let mchapter = chapter;
            if (match[1] && (match2 = match[1].match(/(\d+),/))) {
                mchapter = parseInt(match2[1]) + "章";
            }
            for (let data of chapter_data[mchapter]) {
                if (match[2] === "例") {
                    if (data.dtype === "M" && data.num === match[3]) {
                        break;
                    }
                }
                else if (match[2] === "補足") {
                    if (data.dtype === "S" && data.num === match[3]) {
                        break;
                    }
                }
                else if (match[2] === "ヒント") {
                    if (data.dtype === "H" && data.num === match[3]) {
                        break;
                    }
                }
                else if (match[2] === "章末問題") {
                    if (data.dtype === "Q" && data.num === match[3]) {
                        break;
                    }
                } else if (match[2] === "解答例") {
                    if (data.dtype === "A" && data.num === match[3]) {
                        break;
                    }
                }
                else {
                    if (data.type === match[2] && data.dtype === "" && data.num === match[3]) {
                        break;
                    }
                }
                num++;
            }
            fdata = fdata.replace(re, `!linkdata${linkdata.length}!`);
            if (match[3] === "0") {
                match[3] = "";
            }
            let txt = `[${match[2]}${match[3]}]`
            if (match[4]) {
                match2 = match[4].match(/,(.*)/);
                if (match2[1]) {
                    txt = match2[1];
                }
                else if (match[2] === "補足") {
                    let fname2 = this.calc_fname(chapter_data[mchapter][num].description, mchapter)
                    let fdata2 = filedata[fname2];
                    let match3 = fdata2.match(/<h1>(.*)<\/h1>/);
                    txt = (match3 && match3[1]) || "";
                    txt = txt.replace(/!!supnum!!\s+/, "");
                }
            }
            linkdata.push(`<a href="#" onclick="change_page('${mchapter}', '${num}')" class="material">${txt}</a>`)
        }

        // 置換しないものを保護する
        while (match = fdata.match(/!x([^!]+)x!/)) {
            fdata = fdata.replace(/!x([^!]+)x!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`${match[1]}`);
        }
        // pre タグの中身を保護する
        while (match = fdata.match(/<pre([^<]+)<\/pre>/m)) {
            fdata = fdata.replace(/<pre([^<]+)<\/pre>/m, `!linkdata${linkdata.length}!`);
            linkdata.push(`<pre${match[1]}</pre>`);
        }

        if (this.param.appendix.indexOf("index") < 0) {
            for (const replace_data of replace_data_list) {
                if (this.param.appendix === replace_data.url + ".html") {
                    continue;
                }
                if (replace_data.ignore && this.param.appendix.indexOf(replace_data.ignore) >= 0) {
                    continue;
                }
                if (replace_data.only && this.param.appendix !== replace_data.only + ".html") {
                    continue;
                }
                replace_data.regexp = replace_data.regexp || `()(${replace_data.name})()`;
                const re = new RegExp(replace_data.regexp, "g");
                if (replace_data.url) {
                    fdata = fdata.replace(re, `$1<a href="#" onclick="change_appendix('${replace_data.url}')">$2</a>$3`);
                }
                else {
                    let num = 0;
                    for (let data of chapter_data[replace_data.chapter + "章"]) {
                        if (data.dtype === replace_data.dtype && data.num === replace_data.num) {
                            break;
                        }
                        num++;
                    }
                    fdata = fdata.replace(re, `$1<a href="#" onclick="change_page('${replace_data.chapter}章', '${num}')" class="material">$2</a>$3`);
                }
            }
        }

        let count = 0;
        for (let ldata of linkdata) {
            fdata = fdata.replace(`!linkdata${count}!`, ldata);
            count++;
        }
        //for (let ldata of linkdata2) {
        //    fdata = fdata.replace(/!linkdata2!/, ldata);
        //}
        return fdata;
    }

    create_search() {
        let search = this.param.search;
        let stext = this.escape_text(search.text);
        if (!search.hidden) {
            $("#search_button").addClass("selected");
            $("#search").removeClass("hidden");
            $("#search_result").addClass("hidden");
            $("#search_option").addClass("hidden");
            $("#search_option_button").removeClass("selected");
            this.set_search_class();
            if (stext) {
                let title = `<h1>${stext} の検索結果 ${search.count[0]}件見つかりました。</h1>`;
                if (search.count[0] > 0) {
                    let isfirst = true;
                    const name = ["教材の説明文", "教材のプログラム", "資料"];
                    const id = ["search_material_description_result", "search_material_code_result", "search_appendix_result"];
                    title += "<h2>";
                    for (let i = 1; i <= 3; i++) {
                        if (!isfirst) {
                            title += " ";
                        }
                        if (search.count[i] > 0) {
                            title += `<a href="#" onclick="move_search_id('${id[i - 1]}')">${name[i - 1]} ${search.count[i]}件</a> `;
                            isfirst = false;
                        }
                    }
                    title += "</h2>";
                }
                $("#search_title").html(`${title}`);
            }
            if (!search.option) {
                $("#search_result").removeClass("hidden");
                $("#search_result").html(search.result);
                if (!stext) {
                    $("#search_title").html("<h1>テキストボックスに検索ワードを入れて下さい</h1>");
                }

                let search_title_frame_height = $("#search_header_frame").outerHeight();
                let contents_padding = 10;
                let search_result_frame_height = $("#search_result_frame").outerHeight();
                let search_result_height = $("#search_result").get(0).scrollHeight;
                let max_height = $(window).height() - $("#chapter").height() - 30;
                if (search_title_frame_height + search_result_height > max_height) {
                    $("#search_result").addClass("scrolly");
                    $("#search_result").height(max_height - search_title_frame_height - contents_padding);
                    $("#search").height(max_height)
                }
                else {
                    $("#search_result").removeClass("scrolly");
                    $("#search_result").height("auto");
                    //                    $("#search").height("auto");
                    $("#search").height(search_title_frame_height + search_result_frame_height);
                }
            }
            else {
                $("#search_option_button").addClass("selected");
                $("#search_option").removeClass("hidden");
                $("#search_title").html(`<h1>検索設定</h1>`);
                $("#search").height("auto");
            }
        }
        else {
            $("#search_button").removeClass("selected");
            $("#search").addClass("hidden");
        }
    }

    search() {
        let search = this.param.search;
        let text = search.text = $("#search_text").val();
        text = text.replace(/　/g, " ");
        if (text.trim() === "") {
            return;
        }
        this.param.search.option = false;
        let keywords = text.split(" ");
        let result1 = "<table>";
        let result2 = "<table>";
        let count1 = 0;
        let count2 = 0;
        let searched_code = {};
        if (search.material) {
            for (let chapter = 2; chapter <= 15; chapter++) {
                if (!search.material_chapter[chapter]) {
                    continue;
                }
                let chapter_name = chapter + "章";
                let num = -1;
                for (let data of chapter_data[chapter_name]) {
                    num++;
                    if (!((search.material_type.main && data.dtype === "") ||
                        (search.material_type.misc && data.dtype === "M") ||
                        (search.material_type.sup && data.dtype === "S") ||
                        (search.material_type.question && (data.dtype === "Q" || data.dtype === "H" || data.dtype === "DH" || data.dtype === "A")))) {
                        continue;
                    }

                    if (data.description) {
                        let fname = this.calc_fname(data.description, chapter_name);
                        let fdata = filedata[fname];
                        fdata = this.calc_description(data, chapter_name);
                        if (search.material_description !== "none") {
                            if (search.material_description === "heading") {
                                let match = fdata.match(/<h1>(.*)<\/h1>/);
                                if (match) {
                                    fdata = match[1];
                                }
                                else {
                                    fdata = "";
                                }
                            }
                            let text = this.search_keywords(fdata, keywords);
                            if (text) {
                                count1++;
                                result1 += `<tr><td class="nowrap align-top padding-r10"><a href="#" onclick="change_page('${chapter_name}', ${num})">${chapter_name} ${this.calc_name(data)}</a></td><td>${text}</td></tr>`;
                            }
                        }
                    }
                    for (let [type, typename] of [["html", "HTML"], ["css", "CSS"], ["script", "JS"]]) {
                        if (data[type]) {
                            if (search.material_code === "none" ||
                                (search.material_code !== "all" && search.material_code !== type)) {
                                continue;
                            }
                            let fname = this.calc_fname(data[type], chapter_name);
                            if (searched_code[fname]) {
                                continue;
                            }
                            let fdata = filedata[fname];
                            let text = this.search_code(fdata, keywords);
                            if (text) {
                                count2++;
                                //result2 += `<tr><td class="nowrap align-top padding-r10"><a href="#" onclick="change_page('${chapter_name}', ${num}, '${typename}', '${keywords[0]}')">${chapter_name} ${this.calc_name(data)}</a></td><td class="padding-r10">${typename}</td><td>${text}</td></tr>`;
                                // keyword はもう使っていない？
                                result2 += `<tr><td class="nowrap align-top padding-r10"><a href="#" onclick="change_page('${chapter_name}', ${num}, '${typename}')">${chapter_name} ${this.calc_name(data)}</a></td><td class="padding-r10">${typename}</td><td>${text}</td></tr>`;
                                searched_code[fname] = true;
                            }
                        }
                    }
                }
            }
        }
        result1 += "</table>";
        result2 += "</table>";
        let result3 = "<table>";
        let count3 = 0;
        if (search.appendix !== "none") {
            for (let fname in filedata) {
                if (fname.indexOf("..\\資料\\") < 0) {
                    continue;
                }
                let fdata = filedata[fname];
                fdata = this.calc_appendix(fdata);
                fname = fname.match(/..\\資料\\(.+)\.html/)[1];
                let text = this.search_keywords(fdata, keywords);
                let title = fdata.match(/<h1>(.*)<\/h1>/)[1];
                title = title.replace(/<[^>]*>/g, "");
                let match = title.match(/.+＞([^＞]*)/);
                if (match) {
                    title = match[1];
                }
                if (search.appendix === "heading") {
                    text = this.search_keywords(title, keywords);
                }
                if (text) {
                    count3++;
                    result3 += `<tr><td class="nowrap align-top padding-r10"><a href="#" onclick="change_appendix('${fname}')">${title}</a></td><td>${text}</td></tr>`;
                }
            }
        }
        if (count1 > 0) {
            result1 = `<h2 id='search_material_description_result'>教材の説明文の検索結果（${count1}件）</h2>${result1}`;
        }
        else {
            result1 = "";
        }
        if (count2 > 0) {
            result2 = `<h2 id='search_material_code_result'>教材のプログラムの検索結果（${count2}件）</h2>${result2}`;
        }
        else {
            result2 = "";
        }
        if (count3 > 0) {
            result3 = `<h2 id='search_appendix_result'>資料の検索結果（${count3}件）</h2>${result3}`;
        }
        else {
            result3 = "";
        }
        this.param.search.result = result1 + result2 + result3;;
        this.param.search.count = [count1 + count2 + count3, count1, count2, count3];
        this.redraw(false);
    }

    search_keywords(text, keywords) {
        if (!text) {
            return false;
        }
        let match;
        text = text.replace(/<[^>]*>/g, "");
        text = text.replace(/\s+/g, "");
        let minoffset = Infinity;
        let maxoffset = -Infinity;
        let letterplace = [];
        for (let keyword of keywords) {
            if (!keyword) {
                continue;
            }
            let offset = text.indexOf(keyword);
            if (offset < 0) {
                return false;
            }
            minoffset = Math.min(minoffset, offset);
            maxoffset = Math.max(maxoffset, offset + keyword.length);
            for (let i = 0; i < keyword.length; i++) {
                letterplace[offset + i] = true;
            }
        }
        minoffset = Math.max(0, minoffset - 10);
        maxoffset = Math.min(text.length, maxoffset + 10);
        let retval = "";
        let iskeyword = false;
        if (minoffset > 0) {
            retval += "...";
        }
        let start = minoffset;
        let retval_start = retval.length;
        for (let i = minoffset; i < maxoffset; i++) {
            if (letterplace[i]) {
                if (!iskeyword) {
                    if (i - start > 20) {
                        retval = retval.substr(0, retval_start + 10) + " ... " + retval.substr(-10, 10);
                    }
                    retval += "<span class='em'>";
                }
                iskeyword = true;
            }
            else {
                if (iskeyword) {
                    start = i;
                    retval += "</span>";
                    retval_start = retval.length;
                }
                iskeyword = false;
            }
            retval += text[i];
        }
        if (iskeyword) {
            retval += "</span>";
        }
        if (maxoffset < text.length) {
            retval += "...";
        }
        return retval;
    }

    search_code(text, keywords) {
        if (!text) {
            return false;
        }
        let escaped_keywords = [];
        for (let i = 0; i < keywords.length; i++) {
            escaped_keywords[i] = this.escape_text(keywords[i]);
        }
        let retval = "<pre class='search_code'>";
        let lines = text.split("\n");
        let linenum = 0;
        let keyword_found = {};
        for (let line of lines) {
            linenum++;
            line = line.replace(/\s+/g, "");
            line = " " + this.escape_text(line) + " ";
            let letterplace = [];
            let found = false;
            let line_same = this.param.search.material_code_line.same;
            if (line_same) {
                found = true;
                keyword_found = {};
            }
            for (let keyword of escaped_keywords) {
                if (!keyword || keyword_found[keyword]) {
                    continue;
                }
                if (!this.param.search.material_code_word.same || line.match(RegExp(`[^a-zA-Z0-9_$]${this.escape_regtxt(keyword)}[^a-zA-Z0-9_$]`))) {
                    let offset = line.indexOf(keyword);
                    if (offset < 0) {
                        if (line_same) {
                            found = false;
                        }
                        continue;
                    }
                    for (let i = 0; i < keyword.length; i++) {
                        letterplace[offset + i] = true;
                    }
                    if (!line_same) {
                        found = true;
                    }
                    keyword_found[keyword] = true;
                }
                else if (line_same) {
                    found = false;
                }
            }
            if (found) {
                retval += ("   " + linenum).substr(-3, 3) + "行目 ";
                let iskeyword = false;
                let start = 0;
                let retval_start = retval.length;
                for (let i = 0; i < line.length; i++) {
                    if (letterplace[i]) {
                        if (!iskeyword) {
                            retval += "<span class='em'>";
                        }
                        iskeyword = true;
                    }
                    else {
                        if (iskeyword) {
                            start = i;
                            retval += "</span>";
                            retval_start = retval.length;
                        }
                        iskeyword = false;
                    }
                    retval += line[i];
                }
                if (iskeyword) {
                    retval += "</span>";
                }
                retval += "\n";
                if (line_same) {
                    break;
                }
            }
        }
        retval = retval.substr(0, retval.length - 1);
        retval += "</pre>";
        for (let keyword of escaped_keywords) {
            if (!keyword_found[keyword]) {
                return false;
            }
        }
        return retval;
    }

    create_config() {
        let config = this.param.config;
        this.set_config_class();
        if (!config.hidden) {
            $("#config_button").addClass("selected");
            $("#config").removeClass("hidden");
        }
        else {
            $("#config_button").removeClass("selected");
            $("#config").addClass("hidden");
        }
    }

    escape_text(text = "") {
        return text.replace(/[&'`"<>]/g, function (match) {
            return {
                '&': '&amp;',
                "'": '&#x27;',
                '`': '&#x60;',
                '"': '&quot;',
                '<': '&lt;',
                '>': '&gt;',
            }[match]
        });
    }

    escape_regtxt(text = "") {
        return text.replace(/[-\/\\^$*+?.()|\[\]{}]/g, '\\$&');
    }

    create_index() {
        $("#index").empty();
        $("#index").append(`<div class="heading" id="index_heading">${this.param.chapter}</div>`);
        let table = $("<table></table>");
        //table.addClass("chapter_table");
        $("#index").append(table);
        for (let data of chapter_data[this.param.chapter]) {
            if (data.dtype === "DH") {
                continue;
            }
            let tr = $("<tr></tr>").addClass("nowrap").addClass("chapter_item");
            table.append(tr);
            let name = this.calc_name(data);
            tr.attr('id', `index_${name}`);
            let origname = name;
            let pname = this.find_and_calc_name(data);
            if (data.type !== "" && data.dtype && data.dtype !== "D" && data.dtype !== "R" && data.dtype !== "QS") {
                name += "(" + data.type.substr(0, 1) + ")";
            }
            let td = $(`<td class="index_name">${name}</td>`);
            if (data.iserror) {
                td.addClass("underline");
            }
            let span;
            if (this.param.left === data.fname) {
                td.addClass("selected2");
            }
            else if (this.param.pfname === data.fname) {
                td.addClass("selected3");
            }
            else if ((this.param.right && this.param.right.replace("d.", ".")) === data.fname) {
                td.addClass("selected");
            }
            if ((this.param.show_execute && (this.param.execute_html.replace("d.", ".") === data.fname || this.param.execute_css.replace("d.", ".") === data.fname || this.param.execute_script.replace("d.", ".") === data.fname)) ||
                (!this.param.show_execute && this.param.rightname === origname)) {
                td.addClass("selected4");
                //if (data.pfname && this.param.right === data.fname && this.param.left === null) {
                //    td.addClass("selected5");
                //}
            }
            if (!data.pfname) {
                td.addClass("no_prev");
            }
            td.click(() => {
                this.change_page(data);
                /*
                if (this.param.right === data.fname) {
                    if (data.pfname && this.param.left === null) {
                        this.param.left = data.pfname;
                        this.param.leftname = pname;
                    }
                    else {
                        this.param.left = null;
                        this.param.leftname = "";
                    }
                    this.param.show_diff_textarea = false;
                    this.param.diff_text = null;
                }
                else {
                    if (data.pfname) {
                        this.param.left = data.pfname;
                        this.param.leftname = pname;
                    }
                    else {
                        this.param.left = null;
                        this.param.leftname = "";
                    }
                    this.param.right = data.fname;
                    this.param.pfname = data.pfname;
                    this.param.data = data;
                    this.param.rightname = name;
                    this.param.show_execute_src = null;
                    this.param.show_diff_textarea = false;
                    this.param.diff_text = null;
                    this.param.description = data.description;
                    this.param.desc = data.desc;
                    this.param.desc_emnum = null;
                    this.param.num = data.num;
                    this.param.desc_history = [];
                    if (this.param.show_execute) {
                        if (this.param.execute_html === data.fname) {
                            this.param.show_execute_src = "HTML";
                        }
                        if (this.param.execute_css === data.fname) {
                            this.param.show_execute_src = "CSS";
                        }
                        if (this.param.execute_script === data.fname) {
                            this.param.show_execute_src = "Javascript";
                        }
                    }
                    if (data.dtype === "I") {
                        this.param.description = "I";
                    }
                }
     
                this.show_execute(data);
                this.redraw();
                */
            });
            let pagetxt = "";
            //if (data.page) {
            //    pagetxt = `${data.page}P`
            //}
            tr.append(td); //.append(`<td class="align-right padding-right-5">${pagetxt}</td>`);
            /*
            td = $('<td></td>');
            if (data.pfname) {
                span = $('<span class="button">比</span>');
                span.click(() => {
                    if (this.param.right === data.fname && this.param.left === data.pfname) {
                        this.param.left = null;
                        this.param.leftname = ""
                        this.param.diff_text = null;
                    }
                    else {
                        this.param.left = data.pfname;
                        this.param.leftname = pname;
                        this.param.right = data.fname;
                        this.param.rightname = name;
                        this.param.show_execute_src = null;
                        this.param.diff_text = null;
                        this.param.description = data.description;
                        this.param.desc = data.desc;
                        this.param.desc_emnum = null;
                        this.param.num = data.num;
                        this.param.desc_history = [];
                    }
                    this.show_execute(data);
                    this.redraw();
                })
                if (this.param.right === data.fname && this.param.left !== null) {
                    span.addClass("selected");
                }
                td.append(span);
                
            }
            //else {
            //    if (this.param.left === data.fname) {
            //        td.addClass("align-center");
            //        td.append($('<span>●</span>'));
            //    }
            //}
            tr.append(td);
            */
            /*
            td = $('<td></td>');
            let classname = "button";
            //if (data.pfname) {
            //    classname += " underline";
            //}
            span = $(`<span class="${classname}">表</span>`);
            span.click(() => {
                if (this.param.right === data.fname) {
                    if (data.pfname && this.param.left === null) {
                        this.param.left = data.pfname;
                        this.param.leftname = pname;
                    }
                    else {
                        this.param.left = null;
                        this.param.leftname = "";
                    }
                    this.param.show_diff_textarea = false;
                    this.param.diff_text = null;
                }
                else {
                    if (data.pfname) {
                        this.param.left = data.pfname;
                        this.param.leftname = pname;
                    }
                    else {
                        this.param.left = null;
                        this.param.leftname = "";
                    }
                    this.param.right = data.fname;
                    this.param.rightname = name;
                    this.param.show_execute_src = null;
                    this.param.show_diff_textarea = false;
                    this.param.diff_text = null;
                    this.param.description = data.description;
                    this.param.desc = data.desc;
                    this.param.desc_emnum = null;
                    this.param.num = data.num;
                    this.param.desc_history = [];
                    if (this.param.show_execute) {
                        if (this.param.execute_html === data.fname) {
                            this.param.show_execute_src = "HTML";
                        }
                        if (this.param.execute_css === data.fname) {
                            this.param.show_execute_src = "CSS";
                        }
                        if (this.param.execute_script === data.fname) {
                            this.param.show_execute_src = "Javascript";
                        }
                    }
                }
     
                this.show_execute(data);
                this.redraw();
            });
            if (this.param.right === data.fname) {
                span.addClass("selected");
            }
            td.append(span);
            tr.append(td);
            */
            /*
            td = $('<td></td>');
            tr.append(td);
            if (data.html || data.script) {
                span = $('<span class="button">実</span>');
                if (this.param.show_execute) {
                    if (data.fname == this.param.execute_fname) {
                        span.addClass("selected");
                    }
                }
                span.click(() => {
                    //if (this.param.show_execute && this.param.execute_fname === data.fname) {
                    //    this.param.show_execute = false;
                    //    this.param.execute_fname = null;
                    //    this.param.execute_html = null;
                    //    this.redraw();
                    //    return;
                    //}
                    this.param.left = data.pfname;
                    this.param.leftname = pname;
                    this.param.right = data.fname;
                    this.param.rightname = name;
                    this.param.show_execute_src = null;
                    this.param.diff_text = null;
                    this.param.description = data.description;
                    this.param.desc = data.desc;
                    this.param.desc_emnum = null;
                    this.param.num = data.num;
                    this.param.desc_history = [];
     
                    if (data.html) {
                        let fname = this.calc_fname(data.html);
                        let html = filedata[fname];
                        let match = html.match(/<body>(.*)<\/body>/ims);
                        let body = (match && match[1]) || "このプログラムの実行結果は、コンソールに表示されます。";
                        $("#execute").html(body);
     
                        this.param.execute_fname = data.fname;
                        this.param.execute_html = data.html;
                        this.param.execute_script = data.script;
                        this.param.execute_css = data.css;
                        if (data.script) {
                            let fname = this.calc_fname(data.script);
                            let script = filedata[fname];
                            //script = script.replace("window.onload", "window.script_onload");
                            $("#script").html(script);
                        }
                        else {
                            $("#script").html("");
                        }
                        if (data.css) {
                            let fname = this.calc_fname(data.css);
                            let style = filedata[fname];
                            style = style.replace(/(.*{)/g, "#execute $1");
                            style = style.replace("body", "");
                            $("#style").html(style);
                        }
                        else {
                            $("#style").html("");
                        }
                        this.param.show_execute = true;
                        if (data.type === "JS") {
                            if (this.param.right !== this.param.execute_script) {
                                this.param.left = null;
                                this.param.leftname = "";
                                this.param.diff_text = null;
                            }
                            this.param.show_execute_src = "Javascript";
                            this.param.right = this.param.execute_script;
                            this.param.rightname = this.calc_name(data);//data.script;
                        }
                        else if (data.type === "HTML") {
                            if (this.param.right !== this.param.execute_html) {
                                this.param.left = null;
                                this.param.leftname = "";
                                this.param.diff_text = null;
                            }
                            this.param.show_execute_src = "HTML";
                            this.param.right = this.param.execute_html;
                            this.param.rightname = this.calc_name(data);//data.html;
                        }
                        else {
                            if (this.param.right !== this.param.execute_css) {
                                this.param.left = null;
                                this.param.leftname = "";
                                this.param.diff_text = null;
                            }
                            this.param.show_execute_src = "CSS";
                            this.param.right = this.param.execute_css;
                            this.param.rightname = this.calc_name(data);//data.css;
                        }
                        this.save_and_reload();
                    }
                })
                td.append(span);
            }
            else {
                //if (this.param.show_execute && (this.param.execute_html === data.fname || this.param.excecute_css === data.fname || this.param.execute_css === data.fname)) {
                //    td.addClass("align-center");
                //    td.append($('<span>●</span>'));
                // }
            }
            */
        }
    }

    change_page(data, chapter = null, update_history = true, redraw = true) {
        if (chapter) {
            this.param.chapter = chapter;
        }
        let pname = this.find_and_calc_name(data);
        if (this.param.right === data.fname) {
            if (data.pfname && this.param.left === null) {
                this.param.left = data.pfname;
                this.param.leftname = pname;
            }
            else {
                this.param.left = null;
                this.param.leftname = "";
            }
            this.param.show_diff_textarea = false;
            this.param.diff_text = null;
        }
        else {
            if (data.pfname) {
                this.param.left = data.pfname;
                this.param.leftname = pname;
            }
            else {
                this.param.left = null;
                this.param.leftname = "";
            }
            this.param.right = data.fname;
            this.param.pfname = data.pfname;
            this.param.data = data;
            this.param.rightname = this.calc_name(data);
            this.param.show_execute_src = null;
            this.param.show_diff_textarea = false;
            this.param.diff_text = null;
            this.param.description = data.description;
            this.param.desc = data.desc;
            this.param.dtype = data.dtype;
            this.param.dtypeorig = data.dtypeorig;
            this.param.index = data.index;
            this.param.desc_emnum = null;
            this.param.num = data.num;
            this.param.desc_history = [];
            /*
            if (this.param.show_execute) {
                if (this.param.execute_html === data.fname) {
                    this.param.show_execute_src = "HTML";
                }
                if (this.param.execute_css === data.fname) {
                    this.param.show_execute_src = "CSS";
                }
                if (this.param.execute_script === data.fname) {
                    this.param.show_execute_src = "Javascript";
                }
            }*/
            if (data.dtype === "I") {
                this.param.description = "I";
            }
        }

        if (update_history) {
            this.param.page_history_index++;
            this.param.page_history.length = this.param.page_history_index;
            this.param.page_history.push({ chapter: this.param.chapter, data: JSON.stringify(this.param.data) });
            if (this.param.page_history_index > 100) {
                this.param.page_history.shift();
                this.param.page_history_index--;
                this.param.page_history_length--;
            }
        }
        this.hide_appendix();
        this.param.search.hidden = true;
        this.param.config.hidden = true;
        if (redraw) {
            this.show_execute(data);
            this.redraw();
        }
    }

    show_execute(data) {
        if (data.html) {
            let fname = this.calc_fname(data.html);
            let html = filedata[fname];
            let match = html.match(/<body>(.*)<\/body>/ims);
            let body = (match && match[1]) || "このプログラムの実行結果は、コンソールに表示されます。";
            $("#execute").html(body);

            this.param.execute_fname = data.fname;
            this.param.execute_html = data.html;
            this.param.execute_script = data.script;
            this.param.execute_css = data.css;
            if (data.script) {
                let fname = this.calc_fname(data.script);
                let script = filedata[fname];
                //script = script.replace("window.onload", "window.script_onload");
                $("#script").html(script);
            }
            else {
                $("#script").html("");
            }
            if (data.css) {
                let fname = this.calc_fname(data.css);
                let style = filedata[fname];
                style = style.replace(/(.*{)/g, "#execute $1");
                style = style.replace("body", "");
                $("#style").html(style);
            }
            else {
                $("#style").html("");
            }
            this.param.show_execute = true;
            if (data.type === "JS") {
                if (this.param.right !== this.param.execute_script) {
                    this.param.left = null;
                    this.param.leftname = "";
                    this.param.diff_text = null;
                }
                this.param.show_execute_src = "Javascript";
                this.param.right = this.param.execute_script;
                this.param.rightname = this.calc_name(data);//data.script;
            }
            else if (data.type === "HTML") {
                if (this.param.right !== this.param.execute_html) {
                    this.param.left = null;
                    this.param.leftname = "";
                    this.param.diff_text = null;
                }
                this.param.show_execute_src = "HTML";
                this.param.right = this.param.execute_html;
                this.param.rightname = this.calc_name(data);//data.html;
            }
            else {
                if (this.param.right !== this.param.execute_css) {
                    this.param.left = null;
                    this.param.leftname = "";
                    this.param.diff_text = null;
                }
                this.param.show_execute_src = "CSS";
                this.param.right = this.param.execute_css;
                this.param.rightname = this.calc_name(data);//data.css;
            }

            if (this.param.execute_html === data.fname) {
                this.param.show_execute_src = "HTML";
            }
            if (this.param.execute_css === data.fname) {
                this.param.show_execute_src = "CSS";
            }
            if (this.param.execute_script === data.fname) {
                this.param.show_execute_src = "Javascript";
            }

            this.save_and_reload();
        }
        else {
            this.param.show_execute = false;
            this.param.execute_html = null;
            this.param.execute_script = null;
            this.param.execute_css = null;
            this.save_and_reload();
        }
    }

    save_and_reload() {
        this.param.execute = true;
        this.param.index_scrolltop = $("#index").scrollTop();
        this.save_storage();
        location.reload();
    }

    save_storage() {
        localStorage.setItem("JS_Kyouzai", JSON.stringify(this.param));
    }

    load_storage() {
        let storage = localStorage.getItem("JS_Kyouzai");
        if (storage) {
            storage = JSON.parse(storage);
            this.param = storage;
        }
    }

    execute() {
        if (this.param.execute_script) {
            let fname = this.calc_fname(this.param.execute_script);
            let script = filedata[fname];
            //script = script.replace("window.onload", "window.script_onload");
            //$("#script").html("");
            if (this.param.debug) {
                script = script.replace('"use strict";', '"use strict"; debugger;');
                this.param.debug = false;
            }
            $("#script").html(script);
            //eval(script);
        }
        else {
            $("#script").html("");
            //window.script_onload = null;
            window.onload = null;
        }
        if (this.param.execute_html) {
            let fname = this.calc_fname(this.param.execute_html);
            let match = filedata[fname].match(/<body>(.*)<\/body>/ims);
            if (match && match[1]) {
                $("#execute").html(match[1]);
            }
            else {
                $("#execute").html("");
            }

            match = filedata[fname].match(/<script>(.*)<\/script>/ims);
            if (match) {
                if (this.param.debug) {
                    match[1] = match[1].replace("\n", "\tdebugger;\n");
                    this.param.debug = false;
                }
                //Function(match[1])();
                eval(match[1]);
            }

            match = filedata[fname].match(/document.addEventListener\("DOMContentLoaded", function \(\) {\n(.+)\n/);
            if (match) {
                eval(match[1]);
            }
        }
        if (this.param.execute_css) {
            let fname = this.calc_fname(this.param.execute_css);
            let style = filedata[fname];
            style = style.replace(/(.*{)/g, "#execute $1");
            style = style.replace("body", "");
            $("#style").html(style);
        }
        else {
            $("#style").html("");
        }

        //if (window.script_onload) {
        //    window.script_onload();
        //}
        //if (window.onload) {
        //    window.onload();
        //}
    }

    calc_name(data, ispnum = false) {
        let name = data.name || "JS";
        let num;
        if (!ispnum) {
            num = data.num;
            if (num === "0") {
                num = "";
            }
        }
        else {
            num = data.pnum;
            if (num === "0") {
                num = "";
            }
        }
        return `${name}${num}`;
    }

    /**
     * 名前を探して計算する関数
     * @param {String} name 名前を探す 
     * @returns {String} 計算された文字列
     */
    find_and_calc_name(data) {
        let fname = data.pfname;
        for (let cdata of chapter_data[this.param.chapter]) {
            if (cdata.fname === fname) {
                return this.calc_name(cdata);
            }
        }
        return this.calc_name(data, true);
    }

    create_diff() {
        if (!this.param.delete_comment) {
            $("#comment").addClass("selected");
        }
        else {
            $("#comment").removeClass("selected");
        }
        if (this.param.show_diff_textarea) {
            $("#diff_box").addClass("selected");
        }
        else {
            $("#diff_box").removeClass("selected");
        }

        let fname_left = this.calc_fname(this.param.left);
        let fname_right = this.calc_fname(this.param.right);
        let data_left = filedata[fname_left] || "";
        let data_right = filedata[fname_right] || "";
        if (this.param.diff_text !== null) {
            data_left = this.param.diff_text;
        }
        data_left = data_left.replace(/\r/g, "");
        data_right = data_right.replace(/\r/g, "");

        if (this.param.delete_comment) {
            data_left = this.delete_comment(data_left);
            data_right = this.delete_comment(data_right);
            //data_left = data_left.replace(/[ 　\t]+\/\/[^\n]*\n/mg, "\n");
            //data_right = data_right.replace(/[ 　\t]+\/\/[^\n]*\n/mg, "\n");
        }
        data_left = data_left.replace(/[ 　\t]+\n/mg, "\n");
        data_right = data_right.replace(/[ 　\t]+\n/mg, "\n");

        this.param.diff_context = parseInt(this.param.diff_context) || 4;
        if (this.param.diff_context < 1) {
            this.param.diff_context = 1;
        }
        if (this.param.diff_context > 1000) {
            this.param.diff_context = 1000;
        }

        const unifiedDiff = JsDiff.createPatch("", data_left, data_right, this.param.leftname, this.param.rightname, { context: this.param.diff_context });
        //const unifiedDiff = JsDiff.createPatch("fileName", "a\nb\nc\d\ne\n\nf", "a\nb\nc\d\ne\n\nf", "test", "test");
        const diff2htmlUi = new Diff2HtmlUI({
            diff: unifiedDiff,
        });
        diff2htmlUi.draw('#app', {
            inputFormat: 'json',
            matching: 'words',
            outputFormat: 'side-by-side',
            renderNothingWhenEmpty: true,
            //diffMaxChanges: 1000,
            //diffMaxLineLength: 1000,
        });
        $(".d2h-file-list-wrapper").addClass("hidden");
        $(".d2h-tag").addClass("hidden");
        $(".d2h-info").addClass("hidden");
        $(".d2h-file-side-diff").addClass("align-top");
        $(".d2h-file-side-diff").removeClass("no-overflow");
        $("#diff").addClass("hidden");
        // if (this.param.show_execute) {
        //     $(".d2h-file-side-diff").removeClass("select-text");
        // }
        // else {
        //     $(".d2h-file-side-diff").addClass("select-text");
        // }

        $(".d2h-file-header").eq(0).append($(`
        <div class="float-right">
        <span>前後の表示幅</span>
        <span class="width5"></span>
        <span class="button" id="diff_minus" onclick="showCode.param.diff_context = Math.max(1, showCode.param.diff_context- 1); showCode.redraw();">－</span>
        <span><input type="text" size="1" id="diff_num" value=${this.param.diff_context} onchange="showCode.param.diff_context = this.value; showCode.redraw();"></span>
        <span class="button" id="diff_plus" onclick="showCode.param.diff_context = Math.min(showCode.param.diff_context + 1, 1000); showCode.redraw();">＋</span>
        <span class="button" id="diff_default" onclick="showCode.param.diff_context = 4; showCode.redraw();">規定値</span>
        <span class="button" id="diff_max" onclick="showCode.param.diff_context = 1000; showCode.redraw();">全て</span>
        </div>
        `));
        if (this.param.pfname) {
            $("#diff").removeClass("hidden");

        }
        if (this.param.left || this.param.diff_text) {
            $(".d2h-file-side-diff").eq(0).removeClass("hidden");
            $(".d2h-file-side-diff").eq(1).removeClass("width100");
            $(".d2h-code-line-prefix").removeClass("hidden");
            $(".d2h-ins").removeClass("bgwhite");
            $(".d2h-file-header").removeClass("hidden");
            //$(".d2h-info").removeClass("hidden");
        }
        else if (this.param.show_diff_textarea) {
            $(".d2h-file-side-diff").eq(0).removeClass("hidden");
            $(".d2h-file-side-diff").eq(1).removeClass("width100");
            $(".d2h-file-side-diff").eq(0).empty();
            $(".d2h-file-side-diff").eq(0).append($(`
            <div class="padding5">
            <div id="diff_description">
            右のプログラムと比較したいデータを入力して「比較」ボタンをクリックして下さい。<br>
            このウィンドウにファイルをドラッグして比較することもできます。
            </div>
            <textarea rows="20" id="diff_textarea" class="diff_textarea"></textarea>
            <div>
            <span class="button" id="calc_diff">比較</span>
            <span class="button" id="close_diff">閉じる</span>
            </div>
            </div>
            `));
            $(".d2h-code-line-prefix").addClass("hidden");
            $(".d2h-ins").addClass("bgwhite");
            $(".d2h-file-header").addClass("hidden");
            $("#close_diff").click(() => {
                this.param.show_diff_textarea = false;
                this.redraw();
            });
            $("#calc_diff").click(() => {
                this.param.show_diff_textarea = false;
                this.param.diff_text = $("#diff_textarea").val();
                this.param.left = null;
                this.param.leftname = "テキストボックス";
                if (!this.param.diff_text) {
                    alert("テキストボックスの中身が空です。")
                }
                this.redraw();
            });
            $(".d2h-file-side-diff").eq(0).addClass("no-overflow");
        }
        else {
            $(".d2h-file-side-diff").eq(0).addClass("hidden");
            $(".d2h-file-side-diff").eq(1).addClass("width100");
            $(".d2h-code-line-prefix").addClass("hidden");
            $(".d2h-ins").addClass("bgwhite");
            $(".d2h-file-header").addClass("hidden");
            //$(".d2h-info").addClass("hidden");
        }

        this.diff_set_scrollbar();
        $("#diff_l_sbar_contents").height(1);
        $("#diff_l_sbar").get(0).scrollLeft = 0;
        $("#diff_l_sbar").get(0).onscroll = (evt) => {
            $(".d2h-file-side-diff").get(0).scrollLeft = $("#diff_l_sbar").get(0).scrollLeft;
        };
        $("#diff_r_sbar_contents").height(1);
        $("#diff_r_sbar").get(0).scrollLeft = 0;
        $("#diff_r_sbar").get(0).onscroll = (evt) => {
            $(".d2h-file-side-diff").get(1).scrollLeft = $("#diff_r_sbar").get(0).scrollLeft;
        };
    }

    diff_set_scrollbar() {
        this.param.no_scrollbar = true;
        if ($(".d2h-file-side-diff").get(0)) {
            let offsetwidth_l = $(".d2h-file-side-diff").get(0).offsetWidth;
            $("#diff_l_sbar").width(offsetwidth_l);
            let scrollwidth_l = $(".d2h-file-side-diff").get(0).scrollWidth;
            $("#diff_l_sbar_contents").width(scrollwidth_l);
            if (offsetwidth_l !== scrollwidth_l) {
                this.param.no_scrollbar = false;
            }
        }
        else {
            this.param.no_scrollbar = false;
        }
        if ($(".d2h-file-side-diff").get(1)) {
            let offsetwidth_r = $(".d2h-file-side-diff").get(1).offsetWidth;
            $("#diff_r_sbar").width(offsetwidth_r);
            let scrollwidth_r = $(".d2h-file-side-diff").get(1).scrollWidth;
            $("#diff_r_sbar_contents").width(scrollwidth_r);
            if (offsetwidth_r !== scrollwidth_r) {
                this.param.no_scrollbar = false;
            }
        }
        else {
            this.param.no_scrollbar = false;
        }
        if (this.param.no_scrollbar) {
            $("#diff_l_sbar").addClass("hidden");
            $("#diff_r_sbar").addClass("hidden");
        }
        else {
            $("#diff_l_sbar").removeClass("hidden");
            $("#diff_r_sbar").removeClass("hidden");
        }
    }

    /*
    diff_mousedown(evt) {
        this.param.diff_mousedown = true;
        this.param.mousepos = { x: evt.offsetX, y: evt.offsetY };
    }

    diff_mouseup(evt) {
        this.param.diff_mousedown = false;
    }

    diff_mousemove(evt) {
        if (this.param.diff_mousedown) {
            let diffx = evt.offsetX - this.param.mousepos.x;
            let diffy = evt.offsetY - this.param.mousepos.y;
            console.log($(".d2h-file-side-diff").get(0).scrollLeft);
            $(".d2h-file-side-diff").get(0).scrollLeft -= diffx;//($(".d2h-file-side-diff").get(0).scrollLeft() + diffx);
            this.param.mousepos = { x: evt.offsetX, y: evt.offsetY };
        }
    }*/

    process_diff_files(files) {
        if (files.length > 1) {
            return alert("ファイルは1つだけドラッグアンドドロップして下さい。");
        }
        let file = files[0];
        if (file.type.indexOf("text") !== 0) {
            return alert("テキストファイルではありません。");
        }
        let reader = new FileReader();
        reader.onload = () => {
            if (reader.readyState === 2) {
                this.param.diff_text = reader.result;
                this.param.show_diff_textarea = false;
                this.param.left = null;
                this.param.leftname = file.fname;
                this.redraw();
            }
        };
        reader.readAsText(file);
    }

    set_execute_style() {
        $("#HTML_selector").addClass("hidden");
        $("#CSS_selector").addClass("hidden");
        $("#Javascript_selector").addClass("hidden");
        $("#execute").addClass("hidden");
        $("#start_execute").addClass("hidden");
        $("#debug").addClass("hidden");
        $("#stop_execute").addClass("hidden");
        $("#execute_desc").addClass("hidden");
        $("#execute_minmax").addClass("hidden");
        $("#show_execute").addClass("hidden");

        if (this.param.show_execute) {
            $("#execute").removeClass("hidden");
            $("#execute_minmax").removeClass("hidden");
            $("#show_execute").removeClass("hidden");
            $("#start_execute").removeClass("hidden");
            $("#debug").removeClass("hidden");
            $("#stop_execute").removeClass("hidden");
            if (this.param.execute_html) {
                $("#HTML_selector").removeClass("hidden");
            }
            if (this.param.execute_css) {
                $("#CSS_selector").removeClass("hidden");
            }
            if (this.param.execute_script) {
                $("#Javascript_selector").removeClass("hidden");
            }
            $("#HTML_selector").removeClass("selected");
            $("#CSS_selector").removeClass("selected");
            $("#Javascript_selector").removeClass("selected");
            switch (this.param.show_execute_src) {
                case "HTML":
                    $("#HTML_selector").addClass("selected");
                    break;
                case "CSS":
                    $("#CSS_selector").addClass("selected");
                    break;
                case "Javascript":
                    $("#Javascript_selector").addClass("selected");
                    break;
            }

            let execute_html = false;
            if (this.param.execute_html) {
                let fname = this.calc_fname(this.param.execute_html);
                let match = filedata[fname].match(/<script>(.*)<\/script>/ims);
                if (match) {
                    execute_html = true;
                }
            }
            if (execute_html) {
                if (this.param.execute_min) {
                    $("#show_execute").removeClass("hidden");
                }
                else {
                    $("#execute_minmax").removeClass("hidden");
                    $("#execute_minmax").html("－");
                    $("#show_execute").addClass("hidden");
                    $("#execute").addClass("hidden");
                    $("#execute_desc").removeClass("hidden");
                    $("#execute_desc").html("このプログラムの実行結果は、コンソールに表示されます。");
                }
            }
            else {
                if (this.param.execute_min) {
                    $("#show_execute").removeClass("hidden");
                    //$("#execute_minmax").html("＋");
                    //$("#execute_desc").removeClass("hidden");
                    //$("#execute_html").removeClass("hidden");
                    //$("#execute").addClass("hidden");
                    //$("#execute_desc").html("実行結果を表示するには、右の＋ボタンを押してください。");
                }
                else {
                    $("#show_execute").addClass("hidden");
                    $("#execute_minmax").html("－");
                    $("#execute_desc").addClass("hidden");
                    $("#execute").removeClass("hidden");
                }
            }
        }
        if (this.param.show_execute) {
            if (this.param.execute_min) {//} || this.param.right) {
                $("#exec_container").addClass("hidden");
            }
            else {
                $("#exec_container").removeClass("hidden");
            }
            $("#source_selector").removeClass("hidden");
            $("#code").removeClass("hidden");
        }
        else {
            $("#exec_container").addClass("hidden");
            $("#source_selector").addClass("hidden");
            $("#code").addClass("hidden");
        }
        if (!this.param.left && this.param.pfname) {
            $("#diff").removeClass("selected");
        }
        else {
            $("#diff").addClass("selected");
        }
    }

    set_index_style() {
        let height = $(window).height() - $("#chapter").height() - 25;
        if ($("#index").height() !== height) {
            $("#index").height(height);
        }
        let cheight = $("#index").get(0).scrollHeight;
        if (cheight > height) {
            $("#index").addClass("scrolly");
        }
        else {
            $("#index").removeClass("scrolly");
        }
        //height = $("#execute").height();
        //if (height < 22) {
        //    $("#execute").height(22);
        //}
        let over_height = $("#over_code").height();// + $(".d2h-file-header").get(0).height();
        let file_header_height = 0;
        if ($(".d2h-file-header").eq(0).is(':visible')) {
            file_header_height += $(".d2h-file-header").eq(0).innerHeight();
        }
        height = $(window).height() - $("#chapter").height() - over_height - 46 - file_header_height;
        if (this.param.no_scrollbar) {
            height += 18;
        }
        //cheight = $("#code").get(0).scrollHeight;
        //cheight = $("#app").get(0).scrollHeight;
        cheight = 0;//file_header_height;
        if ($(".d2h-files-diff").eq(0).height()) {
            cheight += $(".d2h-files-diff").get(0).scrollHeight;
        }
        if (cheight > height) {
            //$("#code").height(height);
            //            $("#code").addClass("scrolly");
            $(".d2h-files-diff").eq(0).height(height);
            $(".d2h-files-diff").addClass("scrolly");
        }
        else {
            //$("#code").height(cheight);
            //           $("#code").removeClass("scrolly");
            //$(".d2h-files-diff").eq(0).height(cheight);
            $(".d2h-files-diff").removeClass("scrolly");
        }
        $(".d2h-code-side-linenumber").addClass("relative");
    }

    set_description_style() {
        $("#description_frame").addClass("hidden");
        $("#desc_minmax").addClass("hidden");
        $("#show_description").addClass("hidden");
        $("#desc_prev").addClass("invisible");
        $("#page_prev").addClass("invisible");
        $("#page_next").addClass("invisible");
        $("#description").removeClass("history");
        $("#description_frame").removeClass("history");
        $("#description").css("height", "auto");
        if (this.param.page_history_index > 0) {
            $("#page_prev").removeClass("invisible");
        }
        if (this.param.page_history_index < this.param.page_history.length - 1) {
            $("#page_next").removeClass("invisible");
        }
        if (this.param.description) {
            if (!this.param.show_execute_src) {
                this.param.desc_min = false;
                $("#desc_minmax").addClass("hidden");
            }
            else {
                $("#desc_minmax").removeClass("hidden");
            }
            if (this.param.desc_min) {
                $("#show_description").removeClass("hidden");
                //$("#description_title").html("");
                //$("#description").html("説明を表示するには、右の＋ボタンを押して下さい。");
                //$("#desc_minmax").text("＋");
                //let desc_height = $("#description").get(0).scrollHeight;
                //$("#description_frame").height(desc_height + 20);
                //$("#description").height(desc_height);
            }
            else {
                $("#description_frame").removeClass("hidden");
                let fdata = this.calc_description(this.param, this.param.chapter);

                let match = fdata.match(/(<h1>.*<\/h1>)/);
                if (match) {
                    $("#description_title").html(match[1]);
                    fdata = fdata.replace(/<h1>.*<\/h1>/, "");
                }
                else {
                    $("#description_title").html("");
                }

                if (!this.redrawflag) {
                    $("#description").html(fdata);
                }
                else {
                    this.redrawflag = false;
                }
                if (this.param.desc_emnum) {
                    $(`#${this.param.desc_emnum}`).addClass("em");
                }
                $("#desc_minmax").text("－");
                if (this.param.desc_history && this.param.desc_history.length > 0) {
                    $("#desc_prev").removeClass("invisible");
                    $("#description_frame").addClass("history");
                }

                let desc_title_height = $("#description_title").outerHeight();
                //let a = $("#description_title").html();
                if ($("#description_title").html() === "") {
                    $("#description_title").addClass("hidden");
                    desc_title_height = 0;
                }
                else {
                    $("#description_title").removeClass("hidden");
                }
                let description_padding = 20;
                let desc_height = $("#description").get(0).scrollHeight;
                let desc_frame_height = desc_title_height + desc_height;
                //                if (desc_title_height > 0) {
                //                    desc_frame_height += 10;
                //                }

                let min_height = desc_title_height + Math.min(desc_height, 100);
                let max_height = (($(window).height() - $("#chapter").height()) - 28) * parseFloat(this.param.config.description_maxheight) / 100;
                if (!this.param.show_execute) {
                    max_height = ($(window).height() - $("#chapter").height()) - 28;
                }
                if (max_height < min_height) {
                    max_height = min_height;
                }

                let other_height = $("#chapter").height() + $("#exec_container").height() + $("#source_selector").height() + $("#code").get(0).scrollHeight;
                let height_left = $(window).height() - other_height - 28 - 10;
                $("#description").removeClass("scrolly");
                if (desc_frame_height > height_left) {
                    if (desc_frame_height < min_height) {
                        desc_frame_height = min_height;
                        $("#description").addClass("scrolly");
                    }
                    else if (height_left > max_height) {
                        desc_frame_height = height_left;
                        $("#description").addClass("scrolly");
                    }
                    else if (desc_frame_height > max_height) {
                        desc_frame_height = max_height;
                        $("#description").addClass("scrolly");
                    }
                    //desc_frame_height = height_left - 28;
                }
                //console.log($("#code").height(), other_height, height_left, desc_title_height, desc_height, max_height, min_height, desc_frame_height)
                $("#description_frame").height(desc_frame_height);
                $("#description").height(desc_frame_height - desc_title_height - description_padding);
            }
        }
    }

    calc_description(data, chapter) {
        let fdata = "";
        let fname = this.calc_fname(data.description, chapter);
        let linkdata = [];
        //        if (data.description === "I") {
        if (data.dtype === "I") {
            fdata = `<h1>目次</h1><h2>補足目次</h2>
            <table id="description_table" class="a">
            `;
            let supnumlist = [];
            let num = 0;
            for (let data of chapter_data[chapter]) {
                if (data.dtype === "S") {
                    let supnum = data.num;
                    supnum = parseInt(supnum.replace(/\-\d+/, ""));
                    if (!supnumlist[supnum]) {
                        supnumlist[supnum] = true;
                        let name = data.name + supnum;//this.find_and_calc_name(data) + supnum;
                        let fname2 = this.calc_fname(data.description);
                        let fdata2 = filedata[fname2];
                        let match = fdata2.match(/<h1>(.*)<\/h1>/);
                        let title = (match && match[1]) || "";
                        title = title.replace(/!!supnum!!\s+/, "");
                        fdata += `
                      <tr>
                        <td>${name}</td>
                        <td><a href="#" onclick='change_page("${chapter}", ${num})'>${title}</a></td>
                      </tr>
                    `;
                    }
                }
                num++;
            }
            fdata += "</table>"

            fdata += `<h2>ヒント目次</h2>
            <table id="description_table" class="a">
            `;
            supnumlist = [];
            num = 0;
            for (let data of chapter_data[chapter]) {
                if (data.dtype === "H") {
                    let supnum = data.num;
                    supnum = parseInt(supnum.replace(/\-\d+/, ""));
                    if (!supnumlist[supnum]) {
                        supnumlist[supnum] = true;
                        let name = data.name + supnum;//this.find_and_calc_name(data) + supnum;
                        let fname2 = this.calc_fname(data.description);
                        let fdata2 = filedata[fname2];
                        let match = fdata2.match(/<h1>(.*)<\/h1>/);
                        let title = (match && match[1]) || "";
                        title = title.replace(/!!hintnum!!\s+/, "");
                        fdata += `
                      <tr>
                        <td>${name}</td>
                        <td><a href="#" onclick='change_page("${chapter}", ${num})'>${title}</a></td>
                      </tr>
                    `;
                    }
                }
                num++;
            }
            fdata += "</table>"
        }
        else {
            fdata = filedata[fname] || "";
            let replacetxt = "";
            if (data.dtype === "DH") {
                let fname2 = fname.replace(/(desc\d+)\./, "$1d.");
                replacetxt = filedata[fname2] || "";
            }
            else if (data.dtype === "A") {
                let fname2 = fname.replace(/(desc\d+)\./, "$1a.");
                replacetxt = filedata[fname2] || "";
            }
            fdata = fdata.replace(/!!replacetxt!!/, replacetxt);
        }

        fdata = fdata.replace(/!!num!!/g, data.num);
        if (data.desc === "a") {
            fdata = fdata.replace(/^<h1>ヒント(.*<\/h1>)/, "<h1>解答例$1\n<p>以下の説明は、対応するヒントの内容と同じです。</p>");
            fdata = fdata.replace(/!!hintnum!!/g, `解答例${data.num}`);
            fdata = fdata.replace(/!!hintdesc!!/g, "");
            fdata = fdata.replace(/!!nexthint!!/g, "");
        }
        else if (data.desc === "ac") {
            fdata = "<h1>解答例</h1><div>完成したプログラムの解答例は以下の通りです。</div>";
            if (filedata[fname]) {
                fdata += filedata[fname];
            }
        }
        else {
            if (data.dtype === "H") {
                fdata = fdata.replace(/!!hintnum!!/g, `ヒント${data.num}`);
            }
            if (data.dtype === "A") {
                fdata = fdata.replace(/!!hintnum!!/g, `解答例${data.num}`);
            }
            else {
                fdata = fdata.replace(/!!hintnum!!/g, `ヒント${data.num}（詳細版）`);
            }
            fdata = fdata.replace(/!!supnum!!/g, `[補足${data.num}]`);
        }
        if (fdata.match(/!!supnum-1!!/)) {
            let match = data.num.match(/(\d+)-(\d+)/);
            if (match) {
                fdata = fdata.replace(/!!supnum-1!!/g, `<a href="#" onclick="change_page('${chapter}', '${data.index - 1}')" class="material">[補足${match[1]}-${match[2] - 1}]</a>`);
            }
        }

        let hintplus = 0;
        if (fdata.match(/!!hintnum-1!!/)) {
            hintplus = -1;
        }
        if (fdata.match(/!!hintnum\+1!!/)) {
            hintplus = 1;
        }
        if (hintplus !== 0) {
            let newindex = data.index + hintplus;
            if (data.dtype === "DH" && hintplus == -1) {
                newindex -= 1;
            }
            if (chapter_data[chapter][newindex].dtype === "DH") {
                newindex += hintplus;
            }
            let text = "ヒント";
            if (data.dtype === "A") {
                text = "解答例";
            }
            fdata = fdata.replace(/!!hintnum[-|\+]\d!!/g, `<a href="#" onclick="change_page('${chapter}', ${newindex})" class="material">[${text}${+data.num + hintplus}]</a>`);
        }
        // 詳細版ヒント、解答例などへのリンク
        if (fdata.match(/!!hintlink!!/)) {
            let hintnum = parseInt(data.num);
            let normalhintnum = null;
            let detailhintnum = null;
            let answernum = null;
            let num = 0;
            for (let data of chapter_data[chapter]) {
                if (hintnum === parseInt(data.num)) {
                    if (data.dtype === "H") {
                        normalhintnum = num;
                    }
                    else if (data.dtype === "DH") {
                        detailhintnum = num;
                    }
                    else if (data.dtype === "A") {
                        answernum = num;
                    }
                }
                num++;
            }
            let linktxt = "";
            if (data.dtype !== "A") {
                if (normalhintnum !== null && data.dtype !== "H") {
                    linktxt += `<a href="#" onclick="change_page('${chapter}', ${normalhintnum})" class="material">ヒント簡易版</a>　`;
                }
                if (detailhintnum !== null && data.dtype !== "DH") {
                    linktxt += `<a href="#" onclick="change_page('${chapter}', ${detailhintnum})" class="material">ヒント詳細版</a>　`;
                }
                if (answernum !== null) {
                    linktxt += `<a href="#" onclick="change_page('${chapter}', ${answernum})" class="material">解答例</a>　`;
                }
            }
            else {
                linktxt += `<a href="#" onclick="change_page('${chapter}', ${normalhintnum})" class="material">ヒントに戻る</a>　`;
            }
            linktxt = `<div class="margin-tb10">${linktxt}</div>`;
            fdata = fdata.replace(/!!hintlink!!/, `!linkdata${linkdata.length}!`);
            linkdata.push(linktxt);
        }

        fdata = fdata.replace(/!!answernum!!/g, `解答例${data.num}`);
        // ルールの表の名前
        let rulename = "";
        let match = fdata.match(/!!rname=(.+)!!/);
        if (match) {
            rulename = match[1];
            fdata = fdata.replace(/!!rname=.+!!/, "");
        }
        // 表示する名前
        let txt = "ルール ";
        if (fdata.match(/!!question!!/)) {
            txt = "設問 ";
            fdata = fdata.replace(/!!question!!/, "");
        }

        // ルールのリンク
        match = fdata.match(/!!rule\d+(-\d+)?([^!]+)?!!/g);
        if (match) {
            let rules = new Set(match);
            for (let rule of rules) {
                let [_, num, num2, txt2] = rule.match(/!!rule(\d+)(-\d+)?(.+)?!!/);
                if (num === "0") {
                    num = "";
                }
                let rname = rulename;
                if (num2) {
                    if (num) {
                        rname = `${rulename}-${num}`;
                    }
                    else {
                        rname = `${rulename}${num2}`;
                    }
                    num = `${num}${num2}`;
                }
                let dtxt = `${txt}${num}`;
                if (txt2) {
                    dtxt = txt2;
                }
                if (data.description === "I") {
                    fdata = fdata.replace(new RegExp(rule, "g"), dtxt + " ");
                }
                else {
                    while (fdata.match(new RegExp(rule))) {
                        fdata = fdata.replace(new RegExp(rule), `!linkdata${linkdata.length}!`);
                        linkdata.push(`<a href="#" onclick='change_desc("${rname}", "${num}")' class="rulelink">${dtxt}</a> `);
                    }
                }
            }
        }
        // ルールの表
        match = fdata.match(/!!rtable=(.+)!!/);
        if (match) {
            let rdata = "";
            for (let rule of match[1].split(",")) {
                rdata += "<tr>";
                let rmatch = rule.match(/(\d+(-\d+)?[a-z])/g);
                const classname = {
                    y: "i-yes",
                    n: "i-now",
                    p: "i-par",
                    x: "",
                }
                for (let data of rmatch) {
                    let [_, num, num2, type] = data.match(/(\d+)(-\d+)?([a-z])/);
                    let numtxt = num;
                    let rname = rulename;
                    if (num2) {
                        numtxt += num2;
                        txt = "";
                        rname = rulename + "-" + num;
                    }
                    rdata += `<td class="${classname[type]}">!linkdata${linkdata.length}!</td>`;
                    linkdata.push(`<a href="#" onclick='change_desc("${rname}", "${numtxt}")' class="rulelink">${txt}${numtxt}</a>`);
                }
                rdata += "</tr>";
            }
            let tdata = `
<div class="margin-t5">
実装中：<span class="cbox i-now"></span>
途中まで実装済：<span class="cbox i-par"></span>
実装済：<span class="cbox i-yes"></span>
</div>
<table border class="margin-bottom0">
${rdata}
</table>
`;
            fdata = fdata.replace(/!!rtable=.+!!/, tdata);
        }
        if (data.desc_history && data.desc_history.length > 0) {
            // fdata = `<div class="dummy_prev_button"></div>${fdata}`;
        }
        //if (this.param.desc_emnum) {
        //    let regexp = new RegExp(`(id="${this.param.desc_emnum}")`);
        //    fdata = fdata.replace(regexp, "$1 class='em'");
        //}

        if (match = fdata.match(/^\s*<h1>([^>]*)<\/h1>/)) {
            fdata = fdata.replace(/^\s*<h1>([^>]*)<\/h1>/, `<h1>!linkdata${linkdata.length}!</h1>`);
            linkdata.push(match[1]);
        }

        // 補足のリンク
        if (data.dtype === "S") {
            let supnum = data.num;
            let match = supnum.match(/(\d+)-(\d+)/);
            let prevlinktxt = "　　";
            let nextlinktxt = "　　";
            if (match) {
                supnum = parseInt(match[1]);
                let subnum = parseInt(match[2]);
                let maxsubnum = 1;
                let indexnum = [];
                let num = 0;
                for (let data of chapter_data[chapter]) {
                    match = data.num.match(/(\d+)-(\d+)/);
                    if (match && supnum === parseInt(match[1])) {
                        let subnum2 = parseInt(match[2]);
                        indexnum[subnum2] = num;
                        if (maxsubnum < subnum2) {
                            maxsubnum = subnum2;
                        }
                    }
                    num++;
                }
                if (subnum > 1) {
                    prevlinktxt = `!linkdata${linkdata.length}!`;
                    linkdata.push(`<a href="#" onclick="change_page('${chapter}', ${indexnum[subnum - 1]})">前へ</a>`);
                }
                if (subnum < maxsubnum) {
                    nextlinktxt = `!linkdata${linkdata.length}!`;
                    linkdata.push(`<a href="#" onclick="change_page('${chapter}', ${indexnum[subnum + 1]})">次へ</a>`);
                }
                fdata = fdata.replace(/(<h1>.*)(<\/h1>)/, `$1　　${prevlinktxt}　${nextlinktxt}　!linkdata${linkdata.length}!$2`);
                linkdata.push(`<a href="#" onclick='change_page("${chapter}", 0)'>目次</a>`);
            }
            else {
                fdata = fdata.replace(/(<h1>.*)(<\/h1>)/, `$1　　!linkdata${linkdata.length}!$2`);
                linkdata.push(`<a href="#" onclick='change_page("${chapter}", 0)'>目次</a>`);
            }
        }

        // ヒントのリンク
        if (data.dtype === "H" || data.dtype === "DH") {
            let hintnum = parseInt(data.num);
            let prevlinktxt = "　　";
            let nextlinktxt = "　　";
            let maxnum = 1;
            let indexnum = [];
            let num = 0;
            for (let data of chapter_data[chapter]) {
                if (data.dtype === "H") {
                    let hintnum2 = parseInt(data.num);
                    indexnum[hintnum2] = num;
                    if (maxnum < hintnum2) {
                        maxnum = hintnum2;
                    }
                }
                num++;
            }
            if (hintnum > 1) {
                prevlinktxt = `!linkdata${linkdata.length}!`;
                linkdata.push(`<a href="#" onclick="change_page('${chapter}', ${indexnum[hintnum - 1]})">前へ</a>`);
            }
            if (hintnum < maxnum) {
                nextlinktxt = `!linkdata${linkdata.length}!`;
                linkdata.push(`<a href="#" onclick="change_page('${chapter}', ${indexnum[hintnum + 1]})">次へ</a>`);
            }
            fdata = fdata.replace(/(<h1>.*)<\/h1>/, `$1　　${prevlinktxt}　${nextlinktxt}　!linkdata${linkdata.length}!</h1>`);
            linkdata.push(`<a href="#" onclick='change_page("${chapter}", 0)'>目次</a>`);
        }

        while (match = fdata.match(/!!([^!,]+),([^,!]*)!!/)) {
            fdata = fdata.replace(/!!([^!,]+),([^,!]*)!!/, `!linkdata${linkdata.length}!`);
            let txt = match[2];
            if (!txt) {
                let adata = filedata["..\\資料\\" + match[1] + ".html"];
                let match2 = adata.match(/<h1>.+＞([^<＞]+)<\/h1>/);
                if (match2) {
                    txt = match2[1];
                }
            }
            linkdata.push(`<a href="#" onclick="change_appendix('${match[1]}')">${txt}</a>`)
        }

        while (match = fdata.match(/!\?([^!,]+),([^,!]*),([^,!]*),([^,!]+),([^,!]+)!\?/)) {
            let num = 0;
            for (let data of chapter_data[match[1]]) {
                if (data.type === match[2] && data.dtype === match[3] && data.num === match[4]) {
                    break;
                }
                num++;
            }
            fdata = fdata.replace(/!\?([^!,]+),([^,!]*),([^,!]*),([^,!]+),([^,!]+)!\?/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="#" onclick="change_page('${match[1]}', '${num}')" class="material">${match[5]}</a>`)
        }

        // 外部のリンク
        while (match = fdata.match(/!=([^!,]+),([^!,]+)=!/)) {
            fdata = fdata.replace(/!=([^!,]+),([^!,]+)=!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="https://developer.mozilla.org/ja/docs/${match[1]}" class="external" target="blank">${match[2]}</a>`);
        }

        while (match = fdata.match(/!=([^!,]+)=!/)) {
            fdata = fdata.replace(/!=([^!,]+)=!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="https://developer.mozilla.org/ja/docs/${match[1]}" class="external" target="blank">参考ページ</a>`);
        }

        while (match = fdata.match(/!=([^!,]+),([^!,]+)=!/)) {
            fdata = fdata.replace(/!=([^!,]+),([^!,]+)=!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="https://developer.mozilla.org/ja/docs/${match[1]}" class="external" target="blank">${match[2]}</a>`);
        }

        while (match = fdata.match(/!-([^!,]+),([^!,]+)-!/)) {
            fdata = fdata.replace(/!-([^!,]+),([^!,]+)-!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`<a href="https://developer.mozilla.org/ja/docs/${match[1]}" class="external no_underline" target="blank">${match[2]}</a>`);
        }

        let re = new RegExp(/!\?(\d+,)?(HTML|CSS|JS|例|補足|章末問題|ヒント|解答例)([^,!]+)(,[^,!]*)?!\?/);
        while (match = fdata.match(re)) {
            let num = 0;
            let match2;
            let mchapter = chapter;
            if (match[1] && (match2 = match[1].match(/(\d+),/))) {
                mchapter = parseInt(match2[1]) + "章";
            }
            for (let data of chapter_data[mchapter]) {
                if (match[2] === "例") {
                    if (data.dtype === "M" && data.num === match[3]) {
                        break;
                    }
                }
                else if (match[2] === "補足") {
                    if (data.dtype === "S" && data.num === match[3]) {
                        break;
                    }
                }
                else if (match[2] === "ヒント") {
                    if (data.dtype === "H" && data.num === match[3]) {
                        break;
                    }
                }
                else if (match[2] === "章末問題") {
                    if (data.dtype === "Q" && data.num === match[3]) {
                        break;
                    }
                } else if (match[2] === "解答例") {
                    if (data.dtype === "A" && data.num === match[3]) {
                        break;
                    }
                }
                else {
                    if (data.type === match[2] && data.dtype === "" && data.num === match[3]) {
                        break;
                    }
                }
                num++;
            }
            fdata = fdata.replace(re, `!linkdata${linkdata.length}!`);
            if (match[3] === "0") {
                match[3] = "";
            }
            let txt = `[${match[2]}${match[3]}]`
            if (match[4]) {
                match2 = match[4].match(/,(.*)/);
                if (match2[1]) {
                    txt = match2[1];
                }
                else if (match[2] === "補足") {
                    let fname2 = this.calc_fname(chapter_data[mchapter][num].description, mchapter)
                    let fdata2 = filedata[fname2];
                    let match3 = fdata2.match(/<h1>(.*)<\/h1>/);
                    txt = (match3 && match3[1]) || "";
                    txt = txt.replace(/!!supnum!!\s+/, "");
                }
            }
            linkdata.push(`<a href="#" onclick="change_page('${mchapter}', '${num}')" class="material">${txt}</a>`)
        }

        // 置換しないものを保護する
        while (match = fdata.match(/!x([^!]+)x!/)) {
            fdata = fdata.replace(/!x([^!]+)x!/, `!linkdata${linkdata.length}!`);
            linkdata.push(`${match[1]}`);
        }
        // pre タグの中身を保護する
        while (match = fdata.match(/<pre([^<]+)<\/pre>/m)) {
            fdata = fdata.replace(/<pre([^<]+)<\/pre>/m, `!linkdata${linkdata.length}!`);
            linkdata.push(`<pre${match[1]}</pre>`);
        }

        // 目次の場合は置換しない
        if (data.dtype !== "I") {
            for (const replace_data of replace_data_list) {
                replace_data.regexp = replace_data.regexp || `()(${replace_data.name})()`;
                const re = new RegExp(replace_data.regexp, "g");
                if (replace_data.url) {
                    fdata = fdata.replace(re, `$1<a href="#" onclick="change_appendix('${replace_data.url}')">$2</a>$3`);
                }
                else {
                    if (replace_data.chapter + "章" === this.param.chapter && replace_data.dtype === data.dtype && replace_data.num === data.num) {
                        continue;
                    }
                    let num = 0;
                    for (let data of chapter_data[replace_data.chapter + "章"]) {
                        if (data.dtype === replace_data.dtype && data.num === replace_data.num) {
                            break;
                        }
                        num++;
                    }
                    fdata = fdata.replace(re, `$1<a href="#" onclick="change_page('${replace_data.chapter}章', '${num}')" class="material">$2</a>$3`);
                }
            }
        }

        // 最初の見出しが保護されているので 2 回行う必要がある
        for (let i = 0; i < 2; i++) {
            let count = 0;
            for (let ldata of linkdata) {
                fdata = fdata.replace(`!linkdata${count}!`, ldata);
                count++;
            }
        }

        return fdata;
    }

    set_appendix_style() {
        let appendix_title_frame_height = $("#appendix_header_frame").outerHeight();
        let contents_padding = 10;
        let appendix_contents_frame_height = $("#appendix_contents_frame").outerHeight();
        let appendix_contents_height = $("#appendix_contents").get(0).scrollHeight;
        let max_height = $(window).height() - $("#chapter").height() - 30;
        if (appendix_title_frame_height + appendix_contents_height > max_height) {
            $("#appendix_contents").addClass("scrolly");
            $("#appendix_contents").height(max_height - appendix_title_frame_height - contents_padding);
            $("#appendix").height(max_height)
        }
        else {
            $("#appendix_contents").removeClass("scrolly");
            $("#appendix").height(appendix_title_frame_height + appendix_contents_frame_height);
        }
        if (this.param.appendix_scrollTop) {
            $("#appendix_contents").scrollTop(this.param.appendix_scrollTop);
            if ($("#appendix_contents").scrollTop() === this.param.appendix_scrollTop) {
                this.param.appendix_scrollTop = 0;
            }
        }
    }

    delete_comment(txt) {
        let lines = txt.split("\n");
        let retval = "";
        let iscomment = false;
        for (let line of lines) {
            let newline = "";
            let isslash = false;
            let isbracket = false;
            let isescape = false;
            let isasterisk = false;
            for (let l of line) {
                // \ の次の文字は何であれそのまま入れる
                if (isescape) {
                    newline += l;
                    isescape = false;
                }
                // /* によるコメントでない場合
                else if (!iscomment) {
                    // エスケープの場合
                    if (l === "\\") {
                        isescape = true;
                        newline += l;
                    }
                    // / の次の文字の場合
                    else if (isslash) {
                        // // によるコメントなので、残りは飛ばす
                        if (l === "/") {
                            break;
                        }
                        // /* によるコメントの開始
                        else if (l === "*") {
                            iscomment = true;
                        }
                        else {
                            isslash = false;
                            newline += "/" + l;
                        }
                    }
                    // ” などの間の場合
                    else if (isbracket) {
                        if (l === isbracket) {
                            isbracket = false;
                        }
                        newline += l;
                    }
                    // それら以外の場合
                    else {
                        if (l === '"' || l === "'" || l === "`") {
                            isbracket = l;
                            newline += l;
                        }
                        else if (l === "\\") {
                            isescape = true;
                            newline += l;
                        }
                        else if (l === "/") {
                            isslash = true;
                        }
                        else {
                            newline += l;
                        }
                    }
                }
                // コメントの場合
                else {
                    if (isasterisk) {
                        if (l === "/") {
                            iscomment = false;
                        }
                    }
                    if (l === "*" && iscomment) {
                        isasterisk = true;
                    }
                    else {
                        isasterisk = false;
                    }
                }
            }
            // 行の処理の後処理
            if (isbracket !== "`") {
                isbracket = false;
            }
            // 最初の行以外は、改行を付け加える
            if (retval !== "") {
                retval += "\n"
            }
            retval += newline;
        }
        return retval;
    }
}

function change_desc(dname, num = null) {
    showCode.param.desc_history = showCode.param.desc_history || [];
    showCode.param.desc_history.push([showCode.param.description, showCode.param.desc]);
    showCode.param.description = dname + ".html";
    showCode.param.desc = null;
    showCode.param.desc_emnum = num;
    showCode.redraw();
    //    showCode.set_description_style();
    //    showCode.set_index_style();
}

function change_page(chapter, num, iscode = false, keyword = "") {
    let data = chapter_data[chapter][num];
    showCode.param.right = null;
    showCode.param.left = null;
    showCode.param.change_page = {
        name: showCode.calc_name(data),
        iscode: iscode,
        keyword: keyword,
    };
    showCode.change_page(data, chapter);
}

function change_appendix(fname) {
    showCode.param.appendix_hidden = false;
    showCode.param.appendix_scrollTop = 0;
    showCode.param.search.hidden = true;
    showCode.param.config.hidden = true;
    showCode.param.appendix = fname + ".html";
    showCode.param.appendix_history[++showCode.param.appendix_history_index] = showCode.param.appendix;
    showCode.param.appendix_history.length = showCode.param.appendix_history_index + 1;
    showCode.redraw(false);
}

function move_search_id(id) {
    let scrolltop = $("#search_result").scrollTop();
    $("#search_result").scrollTop($(`#${id}`).offset().top + scrolltop - $("#search_result").offset().top);
}